"""
numba seems to make the fitting 10 times faster

module for fitting resonances curves for kinetic inductance detectors.
written by Jordan Wheeler 12/21/16

for example see res_fit.ipynb in this demos directory

To Do
I think the error analysis on the fit_nonlinear_iq_with_err probably needs some work
add in step by step fitting i.e. first amplitude normalization, then cable delay, then i0,q0 subtraction,
     then phase rotation, then the rest of the fit.
need to have fit option that just specifies tau because that never really changes for your cryostat

Change log
JDW 2017-08-17 added in a keyword/function to allow for gain variation "amp_var" to be taken out before fitting
JDW 2017-08-30 added in fitting for magnitude fitting of resonators i.e. not in iq space
JDW 2018-03-05 added more clever function for guessing x0 for fits
JDW 2018-08-23 added more clever guessing for resonators with large phi into guess separate functions
CHW 2022-09-02 PEP-8 formatting, spelling corrections, and minor code cleanup
"""
import inspect

import numpy as np
import scipy.optimize as optimization
import scipy.stats as stats

from submm.KIDs.res.fit_funcs import linear_mag, nonlinear_mag, nonlinear_iq, nonlinear_iq_for_fitter, \
    nonlinear_mag_for_plot, linear_mag_for_plot, so_resonator_cable, so_resonator_cable_for_fitter
from submm.KIDs.res.data_io import Fit, ResSet, NonlinearIQRes, NonlinearMagRes, LinearMagRes, ResonatorCable
from submm.KIDs.res.utils import amplitude_normalization, calc_qc_qi, guess_x0_iq_nonlinear, guess_x0_mag_nonlinear, \
    guess_x0_iq_nonlinear_sep, guess_x0_mag_nonlinear_sep, guess_so_resonator_cable


def bounds_check(x0, bounds):
    lower_bounds = []
    upper_bounds = []
    for x, lb, ub in zip(x0, bounds[0], bounds[1]):
        if x < lb:
            lower_bounds.append(x * 0.9)
        else:
            lower_bounds.append(lb)
        if x > ub:
            upper_bounds.append(x * 1.1)
        else:
            upper_bounds.append(ub)
    return lower_bounds, upper_bounds


def chi_squared(z, z_fit):
    real_fit = np.real(z_fit)
    imag_fit = np.imag(z_fit)
    real_meas = np.real(z)
    imag_meas = np.imag(z)
    obs_delta = np.sqrt((real_fit - real_meas) ** 2.0 + (imag_fit - imag_meas) ** 2.0)
    chi_sq, p_value = stats.chisquare(f_obs=obs_delta)
    return chi_sq, p_value


def fit_nonlinear_iq(f_hz, z, bounds=None, x0: list = None, fr_guess: float = None, tau=None, tau_guess=None,
                     amp_norm: bool = False, verbose: bool = True):
    """Fit a nonlinear IQ with from an S21 sweep.

    Parameters
    ----------
    f_hz : numpy.array
        frequencies Hz
    z : numpy.array
        complex s21
    bounds : tuple, option (default None)
        A 2d tuple of low values bounds[0] the high values bounds[1] to bound the fitting problem.
    x0 : list, optional (default None)
        The initial guesses for all parameters:
        fr_guess  = x0[0]
        Qr_guess  = x0[1]
        amp_guess = x0[2]
        phi_guess = x0[3]
        a_guess   = x0[4]
        i0_guess  = x0[5]
        q0_guess  = x0[6]
        tau_guess = x0[7]
        f0_guess  = x0[8]
        The fit's initial guess can be very important because least squares fitting does not completely search the
        parameter space.
    fr_guess : float, optional (default None)
        The center resonator frequency in Hz. If None, the center frequency is calculated from the data.
        This overrides the value (x0[0]) specified in the x0 parameter list.
    tau : float, optional (default None)
        If not None, this the fitter to use a fixed value for tau (phase delay), speeding up the calculation.
    tau_guess: float, optional (default None)
        Set the initial guess for tau without. This overrides the value (xo[7]) specified in the x0 parameter list.
    amp_norm: bool, optional (default False)
        When True, a normalization is preformed for the amplitude variable. This parameter is useful when the transfer
        function of the cryostat is not flat.
    verbose : bool, optional (default True)
        Uses the print function to display fit results when true, no prints to the console when false.

    Returns
    -------
    fit : Fit
        A Fit NamedTuple containing the fit results.
    """
    if bounds is None:
        # define default bounds
        if verbose:
            print("default bounds used")
        bounds = ([np.min(f_hz), 50, .01, -np.pi, 0, -np.inf, -np.inf, -1.0e-6, np.min(f_hz)],
                  [np.max(f_hz), 200000, 1, np.pi, 5, np.inf, np.inf, 1.0e-6, np.max(f_hz)])
    if x0 is None:
        # define default initial guess
        if verbose:
            print("default initial guess used")
        # fr_guess = x[np.argmin(np.abs(z))]
        # x0 = [fr_guess,10000.,0.5,0,0,np.mean(np.real(z)),np.mean(np.imag(z)),3*10**-7,fr_guess]
        x0 = guess_x0_iq_nonlinear(f_hz, z)
        # print(x0)
    if fr_guess is not None:
        x0[0] = fr_guess
    if tau is None:
        use_given_tau = False
    else:
        use_given_tau = True
    if tau_guess is not None:
        x0[7] = tau_guess
    if amp_norm:
        z = amplitude_normalization(f_hz, z)
    z_stacked = np.hstack((np.real(z), np.imag(z)))
    # map the initial guess to the standard data record
    guess = NonlinearIQRes(*x0)
    if verbose:
        guess.console(label='Guess', print_header=True)
    # bounds check
    bounds = bounds_check(x0, bounds)
    # fitter choices
    if use_given_tau:
        del bounds[0][7]
        del bounds[1][7]
        del x0[7]
        popt, pcov = optimization.curve_fit(
            lambda x_lamb, a, b, c, d, e, f, g, h: nonlinear_iq_for_fitter(x_lamb, a, b, c, d, e, f, g, tau, h), f_hz,
            z_stacked, x0, bounds=bounds)
        popt = np.insert(popt, 7, tau)
        # fill covariance matrix#
        cov = np.ones((pcov.shape[0] + 1, pcov.shape[1] + 1)) * -1
        cov[0:7, 0:7] = pcov[0:7, 0:7]
        cov[8, 8] = pcov[7, 7]
        cov[8, 0:7] = pcov[7, 0:7]
        cov[0:7, 8] = pcov[0:7, 7]
        pcov = cov
    else:
        popt, pcov = optimization.curve_fit(nonlinear_iq_for_fitter, f_hz, z_stacked, x0, bounds=bounds)
    # human-readable results
    fr, Qr, amp, phi, a, i0, q0, tau, f0 = popt
    Qc, Qi = calc_qc_qi(qr=Qr, amp=amp)
    z_fit = nonlinear_iq(f_hz=f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, i0=i0, q0=q0, tau=tau, f0=f0)
    chi_sq, p_value = chi_squared(z=z, z_fit=z_fit)
    result = NonlinearIQRes(fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, i0=i0, q0=q0, tau=tau, f0=f0,
                            chi_sq=chi_sq, p_value=p_value, Qc=Qc, Qi=Qi)
    if verbose:
        result.console(label='Fit', print_header=True)
    # make a packaged result (NamedTuple) to return
    fit = Fit(origin=inspect.currentframe().f_code.co_name, func=nonlinear_iq,
              guess=guess, result=result, popt=popt, pcov=pcov, f_data=f_hz, z_data=z, flags=set())
    return fit


def fit_nonlinear_iq_sep(fine_f_hz, fine_z, gain_f_hz, gain_z,
                         fine_z_err=None, gain_z_err=None, bounds=None, x0=None, amp_norm: bool = False,
                         verbose: bool = True):
    """Same as fit_nonlinear_iq() but takes fine and gain scans separately

    Parameters
    ----------
    fine_f_hz : numpy.array
        frequencies Hz
    fine_z : numpy.array
        complex s21
    gain_f_hz : numpy.array
        frequencies Hz
    gain_z : numpy.array
        complex s21
    fine_z_err : numpy.array, optional (default None)
        The error in the fine scan, expected to bve the same size as fine_z. If either fine_z_err or gain_z_err is None,
        the error is not used.
    gain_z_err : numpy.array, optional (default None)
        The error in the gain scan, expected to bve the same size as gain_z. If either fine_z_err or gain_z_err is None,
        the error is not used.
    bounds : tuple, option (default None)
        A 2d tuple of low values bounds[0] the high values bounds[1] to bound the fitting problem.
    x0 : list, optional (default None)
        The initial guesses for all parameters:
        fr_guess  = x0[0]
        Qr_guess  = x0[1]
        amp_guess = x0[2]
        phi_guess = x0[3]
        a_guess   = x0[4]
        i0_guess  = x0[5]
        q0_guess  = x0[6]
        tau_guess = x0[7]
        f0_guess  = x0[8]
        The fit's initial guess can be very important because least squares fitting does not completely search the
        parameter space.
    amp_norm : bool, optional (default False)
        When True, a normalization is preformed for the amplitude variable. This parameter is useful when the transfer
        function of the cryostat is not flat.
    verbose : bool, optional (default True)
        Uses the print function to display fit results when true, no prints to the console when false.

    Returns
    -------
    fit : Fit
        A Fit NamedTuple containing the fit results.
    """
    if bounds is None:
        # define default bounds
        print("default bounds used")
        bounds = ([np.min(fine_f_hz), 500., .01, -np.pi, 0, -np.inf, -np.inf, 1 * 10 ** -9, np.min(fine_f_hz)],
                  [np.max(fine_f_hz), 1000000, 1, np.pi, 5, np.inf, np.inf, 1 * 10 ** -6, np.max(fine_f_hz)])
    if x0 is None:
        # define default initial guess
        print("default initial guess used")
        # fr_guess = x[np.argmin(np.abs(z))]
        # x0 = [fr_guess,10000.,0.5,0,0,np.mean(np.real(z)),np.mean(np.imag(z)),3*10**-7,fr_guess]
        x0 = guess_x0_iq_nonlinear_sep(fine_f_hz, fine_z, gain_f_hz, gain_z)
    f_hz = np.hstack((fine_f_hz, gain_f_hz))
    z = np.hstack((fine_z, gain_z))
    if amp_norm:
        z = amplitude_normalization(f_hz, z)
    z_stacked = np.hstack((np.real(z), np.imag(z)))
    # map the initial guess to the standard data record
    guess = NonlinearIQRes(*x0)
    if verbose:
        guess.console(label='Guess', print_header=True)
    # bounds check
    bounds = bounds_check(x0, bounds)
    # error
    if fine_z_err is not None and gain_z_err is not None:
        z_err = np.hstack((fine_z_err, gain_z_err))
        z_err_stacked = np.hstack((np.real(z_err), np.imag(z_err)))
        popt, pcov = optimization.curve_fit(nonlinear_iq_for_fitter, f_hz, z_stacked, x0, sigma=z_err_stacked,
                                            bounds=bounds)
        fr, Qr, amp, phi, a, i0, q0, tau, f0 = popt
        fit_result = nonlinear_iq(f_hz=f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, i0=i0, q0=q0, tau=tau, f0=f0)
        # only do it for fine data
        red_chi_sqr = np.sum((np.hstack((np.real(fine_z), np.imag(fine_z))) - np.hstack(
            (np.real(fit_result[0:len(fine_z)]), np.imag(fit_result[0:len(fine_z)])))) ** 2 / np.hstack(
            (np.real(fine_z_err), np.imag(fine_z_err))) ** 2) / (len(fine_z) * 2. - 8.)

    else:
        popt, pcov = optimization.curve_fit(nonlinear_iq_for_fitter, f_hz, z_stacked, x0, bounds=bounds)
        fr, Qr, amp, phi, a, i0, q0, tau, f0 = popt
        red_chi_sqr = None

    Qc, Qi = calc_qc_qi(qr=Qr, amp=amp)
    z_fit = nonlinear_iq(f_hz=f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, i0=i0, q0=q0, tau=tau, f0=f0)
    chi_sq, p_value = chi_squared(z=z, z_fit=z_fit)
    result = NonlinearIQRes(fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, i0=i0, q0=q0, tau=tau, f0=f0,
                            red_chi_sqr=red_chi_sqr, chi_sq=chi_sq, p_value=p_value, Qc=Qc, Qi=Qi)
    if verbose:
        result.console(label='Fit', print_header=True)
    fit = Fit(origin=inspect.currentframe().f_code.co_name, func=nonlinear_iq,
              guess=guess, result=result, popt=popt, pcov=pcov, f_data=f_hz, z_data=z, flags=set())
    return fit


def fit_nonlinear_iq_with_err(f_hz, z, bounds=None, x0=None, amp_norm: bool = False, verbose: bool = True):
    """Same as fit_nonlinear_iq(), but double fits so that it can get error and a proper covariance matrix out.

    Parameters
    ----------
    f_hz : numpy.array
        frequencies Hz
    z : numpy.array
        complex s21
    bounds : tuple, option (default None)
        A 2d tuple of low values bounds[0] the high values bounds[1] to bound the fitting problem.
    x0 : list, optional (default None)
        The initial guesses for all parameters:
        fr_guess  = x0[0]
        Qr_guess  = x0[1]
        amp_guess = x0[2]
        phi_guess = x0[3]
        a_guess   = x0[4]
        i0_guess  = x0[5]
        q0_guess  = x0[6]
        tau_guess = x0[7]
        f0_guess  = x0[8]
        The fit's initial guess can be very important because least squares fitting does not completely search the
        parameter space.
    amp_norm : bool, optional (default False)
        When True, a normalization is preformed for the amplitude variable. This parameter is useful when the transfer
        function of the cryostat is not flat.
    verbose : bool, optional (default True)
        Uses the print function to display fit results when true, no prints to the console when false.

    Returns
    -------
    fit : Fit
        A Fit NamedTuple containing the fit results.
    """
    if bounds is None:
        # define default bounds
        print("default bounds used")
        bounds = ([np.min(f_hz), 2000, .01, -np.pi, 0, -5, -5, 1 * 10 ** -9, np.min(f_hz)],
                  [np.max(f_hz), 200000, 1, np.pi, 5, 5, 5, 1 * 10 ** -6, np.max(f_hz)])
    if x0 is None:
        # define default initial guess
        print("default initial guess used")
        x0 = guess_x0_iq_nonlinear(f_hz, z)
    if amp_norm:
        z = amplitude_normalization(f_hz, z)
    guess = NonlinearIQRes(*x0)
    if verbose:
        guess.console(label='Guess', print_header=True)
    z_stacked = np.hstack((np.real(z), np.imag(z)))
    # bounds check
    bounds = bounds_check(x0, bounds)
    # fit
    popt_first, pcov_first = optimization.curve_fit(nonlinear_iq_for_fitter, f_hz, z_stacked, x0, bounds=bounds)
    fr_first, Qr_first, amp_first, phi_first, a_first, i0_first, q0_first, tau_first, f0_first = popt_first

    fit_result_stacked = nonlinear_iq_for_fitter(f_hz=z, fr=fr_first, Qr=Qr_first, amp=amp_first, phi=phi_first,
                                                 a=a_first, i0=i0_first, q0=q0_first, tau=tau_first, f0=f0_first)
    # get error
    var = np.sum((z_stacked - fit_result_stacked) ** 2) / (z_stacked.shape[0] - 1)
    err = np.ones(z_stacked.shape[0]) * np.sqrt(var)
    # refit
    popt, pcov = optimization.curve_fit(nonlinear_iq_for_fitter, f_hz, z_stacked, x0, err, bounds=bounds)
    fr, Qr, amp, phi, a, i0, q0, tau, f0 = popt
    # make a dictionary to return
    Qc, Qi = calc_qc_qi(qr=Qr, amp=amp)
    z_fit = nonlinear_iq(f_hz=f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, i0=i0, q0=q0, tau=tau, f0=f0)
    chi_sq, p_value = chi_squared(z=z, z_fit=z_fit)
    result = NonlinearIQRes(fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, i0=i0, q0=q0, tau=tau, f0=f0,
                            chi_sq=chi_sq, p_value=p_value, Qc=Qc, Qi=Qi)
    if verbose:
        result.console(label='Fit', print_header=True)
    fit = Fit(origin=inspect.currentframe().f_code.co_name, func=nonlinear_iq,
              guess=guess, result=result, popt=popt, pcov=pcov, f_data=f_hz, z_data=z, flags=set())
    return fit


# function for fitting an iq sweep with the above equation
def fit_nonlinear_mag(f_hz, z, bounds=None, x0=None, verbose=True):
    """
    Parameters
    ----------
    f_hz : numpy.array
        frequencies Hz
    z : numpy.array
        complex s21
    bounds : tuple, option (default None)
        A 2d tuple of low values bounds[0] the high values bounds[1] to bound the fitting problem.
    x0 : list, optional (default None)
        The initial guesses for all parameters:
        fr_guess  = x0[0]
        Qr_guess  = x0[1]
        amp_guess = x0[2]
        phi_guess = x0[3]
        a_guess   = x0[4]
        b0_guess  = x0[5]
        b1_guess  = x0[6]
        flin_guess= x0[7]
        The fit's initial guess can be very important because least squares fitting does not completely search the
        parameter space.
    verbose : bool, optional (default True)
        Uses the print function to display fit results when true, no prints to the console when false.

    Returns
    -------
    fit : Fit
        A Fit NamedTuple containing the fit results.
    """
    if bounds is None:
        print("default bounds used")
        bounds = ([np.min(f_hz), 100, .01, -np.pi, 0, -np.inf, -np.inf, np.min(f_hz)],
                  [np.max(f_hz), 200000, 1, np.pi, 5, np.inf, np.inf, np.max(f_hz)])
    if x0 is None:
        # define default initial guess
        print("default initial guess used")
        # x0 = [fr_guess,10000.,0.5,0,0,np.abs(z[0])**2,np.abs(z[0])**2,fr_guess]
        x0 = guess_x0_mag_nonlinear(f_hz, z, verbose=verbose)

    guess = NonlinearMagRes(*x0)
    if verbose:
        guess.console(label='Guess', print_header=True)
    # bounds check
    bounds = bounds_check(x0, bounds)
    # fit
    popt, pcov = optimization.curve_fit(nonlinear_mag, f_hz, np.abs(z) ** 2, x0, bounds=bounds)
    fr, Qr, amp, phi, a, b0, b1, flin = popt
    Qc, Qi = calc_qc_qi(qr=Qr, amp=amp)
    z_fit = nonlinear_mag_for_plot(f_hz=f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, b0=b0, b1=b1, flin=flin)
    chi_sq, p_value = chi_squared(z=z, z_fit=z_fit)
    result = NonlinearMagRes(fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, b0=b0, b1=b1, flin=flin,
                             chi_sq=chi_sq, p_value=p_value, Qc=Qc, Qi=Qi)
    if verbose:
        result.console(label='Fit', print_header=True)
    fit = Fit(origin=inspect.currentframe().f_code.co_name, func=nonlinear_mag_for_plot,
              guess=guess, result=result, popt=popt, pcov=pcov, f_data=f_hz, z_data=z, flags=set())
    return fit


def fit_linear_mag(f_hz, z, bounds=None, x0=None, verbose=True):
    """
    Parameters
    ----------
    f_hz : numpy.array
        frequencies Hz
    z : numpy.array
        complex s21
    bounds : tuple, option (default None)
        A 2d tuple of low values bounds[0] the high values bounds[1] to bound the fitting problem.
    x0 : list, optional (default None)
        The initial guesses for all parameters:
        fr_guess  = x0[0]
        Qr_guess  = x0[1]
        amp_guess = x0[2]
        phi_guess = x0[3]
        b0_guess  = x0[4]
        The fit's initial guess can be very important because least squares fitting does not completely search the
        parameter space.
    verbose : bool, optional (default True)
        Uses the print function to display fit results when true, no prints to the console when false.


    Returns
    -------
    fit : Fit
        A Fit NamedTuple containing the fit results.
    """
    if bounds is None:
        # define default bounds
        print("default bounds used")
        bounds = ([np.min(f_hz), 100, .01, -np.pi, -np.inf], [np.max(f_hz), 200000, 1, np.pi, np.inf])
    if x0 is None:
        # define default initial guess
        if verbose:
            print("default initial guess used")
        # x0 = [fr_guess,10000.,0.5,0,0,np.abs(z[0])**2,np.abs(z[0])**2,fr_guess]
        x0 = guess_x0_mag_nonlinear(f_hz, z, verbose=verbose)
        x0 = np.delete(x0, [4, 6, 7])
    guess = LinearMagRes(*x0)
    if verbose:
        guess.console(label='Guess', print_header=True)
    # bounds check
    bounds = bounds_check(x0, bounds)
    # fit
    popt, pcov = optimization.curve_fit(linear_mag, f_hz, np.abs(z) ** 2, x0, bounds=bounds)
    # human-readable results
    fr, Qr, amp, phi, b0 = popt
    Qc, Qi = calc_qc_qi(qr=Qr, amp=amp)
    z_fit = linear_mag_for_plot(f_hz=f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, b0=b0)
    chi_sq, p_value = chi_squared(z=z, z_fit=z_fit)
    result = LinearMagRes(fr=fr, Qr=Qr, amp=amp, phi=phi, b0=b0, chi_sq=chi_sq, p_value=p_value, Qc=Qc, Qi=Qi)
    if verbose:
        result.console(label='Fit', print_header=True)
    fit = Fit(origin=inspect.currentframe().f_code.co_name, func=linear_mag_for_plot,
              guess=guess, result=result, popt=popt, pcov=pcov, f_data=f_hz, z_data=z, flags=set())
    return fit


def fit_nonlinear_mag_sep(fine_f_hz, fine_z, gain_f_hz, gain_z,
                          fine_z_err=None, gain_z_err=None, bounds=None, x0=None, verbose=True):
    """Same as, fit_nonlinear_mag(), above but fine and gain scans are provided separately.
    
    Parameters
    ----------
    fine_f_hz : numpy.array
        frequencies Hz
    fine_z : numpy.array
        complex s21
    gain_f_hz : numpy.array
        frequencies Hz
    gain_z : numpy.array
        complex s21
    fine_z_err : numpy.array, optional (default None)
        The error in the fine scan, expected to bve the same size as fine_z. If either fine_z_err or gain_z_err is None,
        the error is not used.
    gain_z_err : numpy.array, optional (default None)
        The error in the gain scan, expected to bve the same size as gain_z. If either fine_z_err or gain_z_err is None,
        the error is not used.
    bounds : tuple, option (default None)
        A 2d tuple of low values bounds[0] the high values bounds[1] to bound the fitting problem.
    x0 : list, optional (default None)
        The initial guesses for all parameters:
        fr_guess   = x0[0]
        Qr_guess   = x0[1]
        amp_guess  = x0[2]
        phi_guess  = x0[3]
        a_guess    = x0[4]
        b0_guess   = x0[5]
        b1_guess   = x0[6]
        flin_guess = x0[7]
        The fit's initial guess can be very important because least squares fitting does not completely search the
        parameter space.
    verbose : bool, optional (default True)
        Uses the print function to display fit results when true, no prints to the console when false.

    Returns
    -------
    fit : Fit
        A Fit NamedTuple containing the fit results.
    """
    if bounds is None:
        # define default bounds
        print("default bounds used")
        bounds = ([np.min(fine_f_hz), 100, .01, -np.pi, 0, -np.inf, -np.inf, np.min(fine_f_hz)],
                  [np.max(fine_f_hz), 1000000, 100, np.pi, 5, np.inf, np.inf, np.max(fine_f_hz)])
    if x0 is None:
        # define default initial guess
        print("default initial guess used")
        x0 = guess_x0_mag_nonlinear_sep(fine_f_hz, fine_z, gain_f_hz, gain_z)

    guess = NonlinearMagRes(*x0)
    if verbose:
        guess.console(label='Guess', print_header=True)
    # stack the scans for curve_fit
    f_hz = np.hstack((fine_f_hz, gain_f_hz))
    z = np.hstack((fine_z, gain_z))
    # bounds check
    bounds = bounds_check(x0, bounds)
    # fit
    if fine_z_err is not None and gain_z_err is not None:
        z_err = np.hstack((fine_z_err, gain_z_err))
        # propagation of errors left out cross term
        z_err = np.sqrt(4 * np.real(z_err) ** 2 * np.real(z) ** 2 + 4 * np.imag(z_err) ** 2 * np.imag(z) ** 2)
        popt, pcov = optimization.curve_fit(nonlinear_mag, f_hz, np.abs(z) ** 2, x0, sigma=z_err, bounds=bounds)
        fr, Qr, amp, phi, a, b0, b1, flin = popt
        fit_result = nonlinear_mag(f_hz=f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, b0=b0, b1=b1, flin=flin)
        red_chi_sqr = np.sum((np.abs(fine_z) ** 2 - fit_result[0:len(fine_z)]) ** 2 / z_err[0:len(fine_z)] ** 2) / (
                len(fine_z) - 7.)
    else:
        popt, pcov = optimization.curve_fit(nonlinear_mag, f_hz, np.abs(z) ** 2, x0, bounds=bounds)
        fr, Qr, amp, phi, a, b0, b1, flin = popt
        red_chi_sqr = None
    z_fit = nonlinear_mag_for_plot(f_hz=fine_f_hz, fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, b0=b0, b1=b1, flin=flin)
    chi_sq, p_value = chi_squared(z=z, z_fit=z_fit)
    # human-readable results
    Qc, Qi = calc_qc_qi(qr=Qr, amp=amp)
    result = NonlinearMagRes(fr=fr, Qr=Qr, amp=amp, phi=phi, a=a, b0=b0, b1=b1, flin=flin,
                             chi_sq=chi_sq, p_value=p_value, red_chi_sqr=red_chi_sqr,  Qc=Qc, Qi=Qi)
    if verbose:
        result.console(label='Fit', print_header=True)
    fit = Fit(origin=inspect.currentframe().f_code.co_name, func=nonlinear_mag_for_plot,
              guess=guess, result=result, popt=popt, pcov=pcov, f_data=f_hz, z_data=z, flags=set())
    return fit


def fit_so_resonator_cable(f_hz, z, bounds=None, x0=None, verbose=True):
    """
    Parameters
    ----------
    f_hz : numpy.array
        frequencies Hz
    z : numpy.array
        complex s21
    bounds : tuple, option (default None)
        A 2d tuple of low values bounds[0] the high values bounds[1] to bound the fitting problem.
    x0 : list, optional (default None)
        The initial guesses for all parameters:
        fr_guess  = x0[0]
        Qr_guess  = x0[1]
        amp_guess = x0[2]
        phi_guess = x0[3]
        b0_guess  = x0[4]
        The fit's initial guess can be very important because least squares fitting does not completely search the
        parameter space.
    verbose : bool, optional (default True)
        Uses the print function to display fit results when true, no prints to the console when false.


    Returns
    -------
    fit : Fit
        A Fit NamedTuple containing the fit results.
    """
    real = np.real(z)
    imag = np.imag(z)
    z_stacked = np.hstack((np.real(z), np.imag(z)))
    if bounds is None or x0 is None:
        x0_default, bounds_default = guess_so_resonator_cable(freqs=f_hz, real=real, imag=imag)
        if bounds is None:
            bounds = bounds_default
        if x0 is None:
            x0 = x0_default

    guess = ResonatorCable(*x0)
    if verbose:
        guess.console(label='Guess', print_header=True)
    # bounds check
    bounds = bounds_check(x0, bounds)
    # fit
    popt, pcov = optimization.curve_fit(so_resonator_cable_for_fitter, f_hz, z_stacked, x0, bounds=bounds)
    # human-readable results
    f_0, Q, Q_e_real, Q_e_imag, delay, phi, f_min, A_mag, A_slope = popt
    z_fit = so_resonator_cable(f=f_hz, f_0=f_0, Q=Q, Q_e_real=Q_e_real, Q_e_imag=Q_e_imag, delay=delay,
                               phi=phi, f_min=f_min, A_mag=A_mag, A_slope=A_slope)
    chi_sq, p_value = chi_squared(z=z, z_fit=z_fit)
    result = ResonatorCable(fr=f_0, Q=Q, Q_e_real=Q_e_real, Q_e_imag=Q_e_imag, delay=delay,
                            phi=phi, f_min=f_min, A_mag=A_mag, A_slope=A_slope,
                            chi_sq=chi_sq, p_value=p_value)
    if verbose:
        result.console(label='Fit', print_header=True)
    fit = Fit(origin=inspect.currentframe().f_code.co_name, func=so_resonator_cable,
              guess=guess, result=result, popt=popt, pcov=pcov, f_data=f_hz, z_data=z, flags=set())
    return fit


def fit_nonlinear_iq_multi(f_hz, z, center_freqs=None, tau: float = None, fit_overlap=0.5, verbose: bool = True):
    """
    wrapper for handling n resonator fits at once
    mostly just a for loop for fitting but also trys to fit in a way that
    better handles collisions by not fitting to close to other resonators
    f_hz and z should have shape n_iq_points x n_res points
    center_freqs can be specified if you are fitting n resonators but you
    know there are actually more the n resonators at the frequency locations
    in center_freqs. This is useful if you didn't collect data for all of the 
    resonators but don't want collisions to screw up your fitting.
    fit_overlap: default 0.5 will only use data to halfway between the resonator
    you are trying to fit and the nearest neighbor resonators if it is close by.
    returns a class
    """

    if center_freqs is None:
        center_freqs = f_hz[f_hz.shape[0] // 2, :]
    res_fits = []
    for i in range(0, f_hz.shape[1]):
        f_single = f_hz[:, i]
        z_single = z[:, i]
        # flag data that is too close to other resonators
        if center_freqs is not None:
            center_index = np.argmin(np.abs(center_freqs - f_single[len(f_single) // 2]))
        else:
            center_index = i
        distance = center_freqs - center_freqs[center_index]
        if center_freqs[center_index] != np.min(center_freqs):  # don't do if lowest frequency resonator
            closest_lower_dist = -np.min(np.abs(distance[np.where(distance < 0)]))
            closest_lower_index = np.where(distance == closest_lower_dist)[0][0]
            halfway_low = center_freqs[center_index] - \
                          (center_freqs[center_index] - center_freqs[closest_lower_index]) * fit_overlap
        else:
            halfway_low = 0

        if center_freqs[center_index] != np.max(center_freqs):  # don't do if highest frequency resonator
            closest_higher_dist = np.min(np.abs(distance[np.where(distance > 0)]))
            closest_higher_index = np.where(distance == closest_higher_dist)[0][0]
            halfway_high = center_freqs[center_index] + \
                           (center_freqs[closest_higher_index] - center_freqs[center_index]) * fit_overlap
        else:
            halfway_high = np.inf

        use_index = np.where(((f_single > halfway_low) & (f_single < halfway_high)))
        mask = np.zeros(len(f_single))
        mask[use_index] = 1
        f_single_res = f_single[use_index]
        z_single_res = z_single[use_index]

        try:
            fit_single_res = fit_nonlinear_iq(f_single_res, z_single_res, tau=tau, verbose=verbose)
        except Exception as e:
            if verbose:
                print(e)
                print("failed to fit")
        else:
            # overwrite fit data with full data set
            fit_dict = fit_single_res._asdict()
            fit_dict['f_data'] = f_single
            fit_dict['z_data'] = z_single
            fit_dict['mask'] = mask
            fit = Fit(**fit_dict)
            res_fits.append(fit)
    res_set = ResSet(res_fits=res_fits, verbose=verbose)
    return res_set


def fit_linear_mag_multi(f_hz, z, verbose: bool = True):
    """
    wrapper for handling n resonator fits at once
    f_hz and z should have shape n_iq_points x n_res points
    return same thing as fitter but in arrays for all resonators
    """
    center_freqs = f_hz[f_hz.shape[0] // 2, :]
    res_fits = []
    for i in range(0, f_hz.shape[1]):
        f_single = f_hz[:, i]
        z_single = z[:, i]
        # flag data that is too close to other resonators              
        distance = center_freqs - center_freqs[i]
        if center_freqs[i] != np.min(center_freqs):  # don't do if lowest frequency resonator
            closest_lower_dist = -np.min(np.abs(distance[np.where(distance < 0)]))
            closest_lower_index = np.where(distance == closest_lower_dist)[0][0]
            halfway_low = (center_freqs[i] + center_freqs[closest_lower_index]) / 2.
        else:
            halfway_low = 0

        if center_freqs[i] != np.max(center_freqs):  # don't do if highest frequency
            closest_higher_dist = np.min(np.abs(distance[np.where(distance > 0)]))
            closest_higher_index = np.where(distance == closest_higher_dist)[0][0]
            halfway_high = (center_freqs[i] + center_freqs[closest_higher_index]) / 2.
        else:
            halfway_high = np.inf
        use_index = np.where(((f_single > halfway_low) & (f_single < halfway_high)))
        mask = np.zeros(len(f_single))
        mask[use_index] = 1
        f_single_res = f_single[use_index]
        z_single_res = z_single[use_index]
        try:
            fit_single_res = fit_linear_mag(f_single_res, z_single_res, verbose=verbose)
        except Exception as e:
            if verbose:
                print("problem")
                print(e)
                print("failed to fit")
        else:
            # overwrite fit data with full data set
            fit_dict = fit_single_res._asdict()
            fit_dict['f_data'] = f_single
            fit_dict['z_data'] = z_single
            fit_dict['mask'] = mask
            fit = Fit(**fit_dict)
            res_fits.append(fit)
    res_set = ResSet(res_fits=res_fits, verbose=verbose)
    return res_set


if __name__ == '__main__':
    # get demo data
    import os
    from scipy.io import loadmat
    from submm.sample_data.abs_paths import abs_path_sample_data

    data_path = os.path.join(abs_path_sample_data,
                             "Survey_Tbb20.000K_Tbath170mK_Pow-60dBm_array_temp_sweep_long.mat")
    sample_data = loadmat(data_path)

    freq_ghz = sample_data['f'][:, 0]
    freq_mhz = freq_ghz * 1.0e3
    freq_hz = freq_ghz * 1.0e9
    s21_complex = sample_data['z'][:, 0]
    s21_mag = 20 * np.log10(np.abs(s21_complex))

    freq_hz_res1 = freq_hz[21050: 21250]
    freq_mhz_res1 = freq_hz[21050: 21250]
    s21_complex_res1 = s21_complex[21050: 21250]
    s21_mag_res1 = s21_mag[21050: 21250]

    # # Caleb's testing area
    # Test the Res() instance for the fit result directly
    res_fit = fit_nonlinear_iq(f_hz=freq_hz_res1, z=s21_complex_res1)
    print(f'Res fit res_fit.result["fr"] from key = {res_fit.result["fr"]}')
    print(f'res_fit.result[0], from index = {res_fit.result[0]}')
    print(f'res_fit.result[0:2], from slice = {res_fit.result[0:2]}')
    print(f'res_fit.result.fr, from attribute = {res_fit.result.fr}\n')

    # test the Fit() instance for to get Res parameters
    print(f'Res fit fr (res_fit["fr"]) from key = {res_fit.result["fr"]}')
    print(f'Res fit Qc (res_fit["Qc"]) from key = {res_fit.result["Qc"]}\n')

    # test the console (print data to screen) output
    res_fit.console()

    # test hash ability of the result
    test_set = {res_fit.result}

    # test the plot
    # res_fit.plot()

    # test write, read, and iteration class
    res_set_test = ResSet(res_fits=[res_fit])
    res_set_test.write()
    res_set_read = ResSet(path=res_set_test.path)
    for read_result in res_set_read:
        read_result.console(label='Read', fields=['fr', 'tau', 'Qc'])
    # get array of the results for specific fields
    res_set_test('fr', 'amp', 'z_data')



import os
import time
import numpy as np
import matplotlib as mpl

from scipy.io import loadmat

from submm.KIDs import find_resonances_interactive as find_kids
from submm.sample_data.abs_paths import abs_path_sample_data
from submm.KIDs.res.sweep_tools import InteractivePlot
from submm.KIDs.res.fitting import fit_nonlinear_iq_multi, fit_linear_mag_multi


def main(linear=False):  # if resonators are well below bifurcation fitting can be much faster

    # load the sample data
    data_path = os.path.join(abs_path_sample_data,
                             "Survey_Tbb20.000K_Tbath170mK_Pow-60dBm_array_temp_sweep_long.mat")
    sample_data = loadmat(data_path)

    # recast the data
    freq_ghz = sample_data['f'][:, 0]
    freq_mhz = freq_ghz * 1.0e3
    freq_hz = freq_ghz * 1.0e9
    s21_complex = sample_data['z'][:, 0]
    s21_mag = 20 * np.log10(np.abs(s21_complex))

    # find resonators interactive
    # 1st filter
    # 2nd find peaks
    # 3rd manual correction
    ip = find_kids.find_vna_sweep(freq_hz, s21_complex)

    # slice up the vna with a span equal to q_slice where q = f/delta_f
    res_freq_array, res_array = find_kids.slice_vna(freq_hz, s21_complex, ip.kid_idx, q_slice=2000, flag_collided=False)

    # fit the resonators
    t1 = time.time()
    if not linear:
        res_set = fit_nonlinear_iq_multi(res_freq_array, res_array, tau=97 * 10 ** -9)
    else:
        res_set = fit_linear_mag_multi(res_freq_array, res_array)
    t2 = time.time()
    print("time to fit {:.2f} s".format(t2 - t1))

    # if you want to inspect just the first fit to understand the data
    result = next(iter(res_set))
    # and then for fit_result
    fit_result = res_set._fit_results[result]
    # to see what is in the class
    print("fields in a result")
    print(result._fields)
    print("fields in a fit_result")
    print(fit_result._fields)

    # plot the data, see submm/KIDs/res/data_io.py's ResSet plot function for example of sweep plottingzz
    ip2 = res_set.plot(flags=ip.flags)

    # below is example of retuning resonators

    # ip2 = InteractivePlot(res_freq_array,res_array,retune = True,find_min = False)


if __name__ == "__main__":
    main()



import sys
import getopt
import time

from submm.lab_brick.core import Attenuator


def attenuation_change(device, status_byte, count, byteblock, val):
	print("User Defined Callback for attenuation_level_change")
	print("Attenuation: " + str(val))


def attenuation_response(device, status_byte, count, byteblock, val):
	print("User Defined Callback for attenuation_level_response")
	print("Attenuation: " + str(val))


if __name__ == '__main__':

	try:
		opts, args = getopt.getopt(sys.argv[1:], "hv:p:", ["help", "vid=", "pid="])
	except getopt.GetoptError as err:
		# print help information and exit:
		print(str(err))  # will print something like "option -a not recognized"
		usage()
		sys.exit(2)
		
	vid = -1
	pid = -1
		
	for o, a in opts:
		if o in ("-h", "--help"):
			usage()
			sys.exit()
		elif o in ("-v", "--vid"):
			vid = a
		elif o in ("-p", "--pid"):
			pid = a
			
	if vid == -1 or pid == -1:
		usage()
		sys.exit()
	else:
		try:
			vid = int(vid, 0)
			pid = int(pid, 0)
		except:
			usage()

		att = Attenuator(vid = vid, pid = pid, debug=True)
		#att = Attenuator(vid = vid, pid = pid, debug=False)

		att.register_callback('attenuation_level_response', attenuation_response)
		
		att.get_attenuation(debug=True)
		
		att.register_callback('attenuation_level_change', attenuation_change)

		#att.set_ramp_start_attenuation_level(20)
		time.sleep(0.1)
		#att.set_ramp_stop_attenuation_level(1)
		#att.set_ramp_step_size(1)
		#att.start_ramp_down()

		att.set_ramp_start_attenuation_level(1, debug=True)
		att.set_ramp_stop_attenuation_level(5, debug=True)

		att.set_ramp_step_size(1, debug=True)
		att.set_ramp_dwell_time(500, debug=True)
		att.set_ramp_wait_time(500, debug=True)

		att.set_ramp_dwell_time_bidirectional(500, debug=True)
		att.set_ramp_wait_time_bidirectional(500, debug=True)
		att.set_ramp_step_size_bidirectional(1, debug=True)

		att.start_ramp_continuous_bidirectional(debug=True)

	while 1:
		#att.set_attenuation(5, debug=True)
		time.sleep(10)
		#att.set_attenuation(10, debug=True)
		#time.sleep(10)



import pyvisa
import time


class attenuator(object):

    # Default constructor used to call the instrument at the given
    # GPIBaddress and initializes the object to the instrument
    def __init__(self,address = None):
        if not address:
            self.GPIBaddress = 'GPIB0::10::INSTR'
        else:
            self.GPIBaddress = address
        rm = pyvisa.ResourceManager("@py")
        self.obj1 = rm.open_resource(self.GPIBaddress)

    def get_attenuation(self, channel=1):
        cmd_str = 'CHAN %d' % (channel)
        self.obj1.write(cmd_str)
        time.sleep(.25)
        attenuation = float(self.obj1.query('ATTN?'))
        return attenuation

    def set_attenuation(self, attenuation, channel=1):

        cmd_str = 'CHAN %d' % (channel)
        self.obj1.write(cmd_str)
        time.sleep(.15)
        cmd_str = 'ATTN %d' % (attenuation)
        self.obj1.write(cmd_str)
        time.sleep(.15)



import numpy as np
import scipy.special as special
import scipy.optimize as optimization
import matplotlib.pyplot as plt

# this is a list of definitions that can be used to predict noise in KIDS
# right now it just contains the nessasary requirements for perdicting G-R noise in TiN
# I should expand it to include some of the definitions in Jason code
# this is adapted from some IDL code I got from Matt Bradford

#Written by Jordan on 12/12/2016

#Change log
# 1/6/2017 added nqp_min as a specified parameter for grnoise 



def nqp(t,tc,v,nqp_min):
    '''
    if ~keyword_set(nqp_min) then nqp_min=400. ; per cubic micron: zero-Temp residual QP density
    V is in cubic microns
    N0=1.72e10 for aluminum
    '''
    N0=4.e10 # for TiN
    N0 = N0/1.6e-19 # now microns^3 / Joule

    #Delta = double (3.5 * 1.381e-23 * tc)
    Delta = 1.74 * 1.381e-23 * tc #factor in delta is suspect but it would be canceled out by a factor in tc
    Nqp = v*2*N0*np.sqrt(2*np.pi*Delta*1.381e-23*t)*np.exp(-1*Delta/(1.381e-23*t))+v*nqp_min

    return Nqp

def deltaf_f(t, tc, nu, alpha, gamma):
    '''
    From Steve Hailey-Dunsheath, 2 Februay 2016
    Calculate a model for the fractional frequency shift due to change in bath temperature
    This is (alpha)*(gamma/2)*[sigma2(T)/sigma2(T=0)-1], where sigma2(T) is
    from equation 2.92 in Gao thesis, and equation 20 in Gao+08 JLTP

    nu is in MHz
    if ~keyword_set(alpha) then alpha=1
    if ~keyword_set(gamma) then gamma=1
    '''
    d_0 = 1.762*tc #factor of 1.762 is suspect
    xi = 6.626e-34*nu*1.e6/(2.*1.38e-23*t)
    model = -1.*alpha*gamma/2.*np.exp(-1.*d_0/t)*((2.*np.pi*t/d_0)**0.5 + 2.*np.exp(-1.*xi)*special.iv(0,xi))
    return model

def deltaf_f_vec(t,tc,nu,alpha,gamma):
    d_0 = 1.762*tc #factor of 1.762 is suspect
    t = np.reshape(t,(t.shape[0],1,1))
    xi = 6.626e-34*nu*1.e6/(2.*1.38e-23*t)
    model = -1.*alpha*gamma/2.*np.exp(-1.*d_0/t)*((2.*np.pi*t/d_0)**0.5 + 2.*np.exp(-1.*xi)*special.iv(0,xi))
    return model


def df_response(t,tc,f):
    '''
    calculate d (df/f) / dT via finite difference
    f is in MHz
    calls deltaf_f which computes frequency shift
    '''
    delta_t = t/100.
    dff_dt = (deltaf_f(t+delta_t,tc,f,1,1) - deltaf_f(t-delta_t,tc,f,1,1))/(2*delta_t)
    return dff_dt


def grnoise(t,tc,V,tau_qp,N0,f,nqp_min):
    '''
    this function calculates gr noise assuming constant tau
    fuction below responsivity probably does a better job. 
    V (volume in cubic microns)
    if ~keyword_set(N0) then N0=4.e10 ; microns^3 / eV
    if ~keyword_set(tau_qp) then tau_qp = 5e-6 ; sec
    if ~keyword_set(nqp_min) then nqp_min=400. ; QP per cubic micron at zero Temp
    '''
    N0 = N0/1.6e-19 # now microns^3 / Joule

    #ef^2 = 4 beta^2 Nqp tau_qp
    #beta = df_0 / d Nqp , so use (df_0/ dT) (dT / dNqp)

    #Delta = double (3.5 * 1.381e-23 * tc)
    Delta = 1.74*1.381e-23*tc
    #Nqp = v * 2 * N0 * sqrt(2*!pi*delta * 1.381e-23 * T) *exp(-1*Delta / (1.381e-23 * T)) + v * nqp_min
    dNqp_dt = V*2.*N0*np.sqrt(2*np.pi*Delta*1.318e-23)*np.exp(-1.*Delta/(1.381e-23*t))*(1./(2.*np.sqrt(t)) + np.sqrt(t)*Delta/1.381e-23/t**2)
    delta_t = t/100.
    dNqp_dt = (nqp(t+delta_t,tc,V,nqp_min)-nqp(t-delta_t,tc,V,nqp_min))/(2*delta_t)

    beta = df_response(t,tc,f)/dNqp_dt
    #assume a frequency of 100 MHz here, shouldn't matter.

    ef2 = 4. * beta**2 * nqp(t,tc,V,nqp_min) * tau_qp
    return ef2


def responsivity(temp,pabs,tc = 1.,N0 = 4.*10**10,nstar =100.,tau_max = 100.,eta_pb = 0.7,vol = 1.,fr = 100.,alpha_k = 1.,gamma_t = 1.,nu_opt = 250, n_gamma = 0.):
    '''
    Special thanks to Steve Hailey Dunsheath whose made this orginal function in idl 
    '''
    #Define various constants
    k_B = 1.381*10**-23 #Boltzmann constant [J K^-1]
    ev_joule = 1.6022*10**-19 #eV/J ratio [eV J^-1]
    h_p = 6.626*10**-34 #Planck constant [J s]

    #Compute the ratio of Delta_0/k_B [K]
    d0_kB = 1.764*tc

    # Compute n thermal
    nth = 2.*N0*k_B/ev_joule*np.sqrt(2.*np.pi*temp*d0_kB)*np.exp(-1.*d0_kB/temp)

    # Compute nqp
    # This expression has a term of the form [sqrt(1 + eps) - 1], where eps is small when nth and pabs are small.
    # When eps is small this term is not computed accurately, and here we explicitly linearize it.
    #nqp = nstar*(-1. + sqrt(1. + 2.*nth/nstar + (nth/nstar)^2. + 2.*eta_pb*pabs*1.e-12*tau_max*1.e-6/(nstar*vol*d0_kB*k_B)))
    eps = 2.*nth/nstar + (nth/nstar)**2. + 2.*eta_pb*pabs*1.*10**-12*tau_max*1.*10**-6/(nstar*vol*d0_kB*k_B)
    term = np.sqrt(1. + eps) - 1.
    indx = np.where(eps < 1*10**-8)
    count = len(indx)
    if (count != 0):
        term[indx] = 0.5*eps[indx]
    nqp = nstar*term
    
    #Compute tau_qp
    tau_qp = tau_max/(1. + nqp/nstar)

    # Compute S1 and S2
    xi = h_p*fr*1*10**6/(2.*k_B*temp)
    s1 = (2./np.pi)*np.sqrt(2.*d0_kB/(np.pi*temp))*np.sinh(xi)*special.kv(0,xi)
    s2 = 1. + np.sqrt(2.*d0_kB/(np.pi*temp))*np.exp(-1.*xi)*special.iv(0,xi)
    #s2 = 3.

    #Compute xr and Qi_inv
    #Note that xr refers to the frequency shift from the nqp = 0 state
    xr = -1.*alpha_k*gamma_t*s2*nqp/(4.*N0*d0_kB*k_B/ev_joule)
    Qi_inv = -1.*xr*2.*s1/s2

    #Compute the frequency and Qinv responsivity
    r_x = -1.*alpha_k*gamma_t*s2/(4.*N0*d0_kB*k_B/ev_joule)*eta_pb*tau_qp*1*10**-6/(d0_kB*k_B*vol)
    r_qinv = -1.*r_x*2.*s1/s2

    #Compute Sxx_gr and Sxx_gr0
    tau_th = tau_max/(1. + nth/nstar) #quasiparticle lifetime for a superconductor in thermal equilibrium at the specified temperature [microsec]
    gamma_th = nth*vol/2.*(1./tau_max + 1./tau_th)*1*10**6 #quasiparticle generation rate due to thermal fluctuations ;[sec^-1]
    gamma_r = nqp*vol/2.*(1./tau_max + 1./tau_qp)*1*10**6 #quasiparticle recombination rate ;[sec^-1]
    sxx_gr = (alpha_k*gamma_t*s2/(4.*N0*d0_kB*k_B/ev_joule))**2.*4.*(tau_qp*1*10**-6)**2./vol**2.*(gamma_th + gamma_r)
    sxx_gr0 = (alpha_k*gamma_t*s2/(4.*N0*d0_kB*k_B/ev_joule))**2.*4.*nqp/vol*tau_qp*1*10**-6

    #Compute Sxx_gamma
    sxx_gamma = (r_x)**2.*2.*h_p*nu_opt*1*10**9*pabs*1*10**-12*(1. + n_gamma)

    #Define the output dictionary and return
    dict = {'nth':nth, 
          'nqp':nqp, 
          'tau_qp':tau_qp, 
          's1':s1, 
          's2':s2, 
          'xr':xr, 
          'Qi_inv':Qi_inv, 
          'r_x':r_x, 
          'r_qinv':r_qinv, 
          'sxx_gr':sxx_gr, 
          'sxx_gr0':sxx_gr0, 
          'sxx_gamma':sxx_gamma}
        
    return dict

def responsivity_help():
    print("The input variables are temp,pabs,tc = 1.,N0 = 4.*10**10,nstar =100.,tau_max = 100.,eta_pb = 0.7,vol = 1.,fr = 100.,alpha_k = 1.,gamma_t = 1.,nu_opt = 250, n_gamma = 0.")
    print("The output variables are nth, nqp, tau_qp, s1, s2, xr, Qi_inv, r_x, r_qinv, sxx_gr, sxx_gr0, sxx_gamma")


def f0dirshort(T, f00, Fdelta):
    '''
    fucntion that computes TLS frequency shift 
     is center frequency in GHz, p2 is the product of filling factor and
    loss tangent Fdelta_TLS, returns the frequency in GHz
    Taken from Jiansong Gao's Matlab code
    '''
    f01K = 20.8366
    ref0 =np.real(special.digamma(1/2 + 1/(2*np.pi*1j)*f00/f01K/T))-np.log(f00/f01K/T/(2*np.pi)); 
    y = f00 + f00*Fdelta*1/np.pi*ref0;
    return y

def fit_tls(T,f,sigma = None,**keywords):
    x0 = np.asarray((f[0],1e-5))
    if sigma is not None:
        print("using error")
        print(sigma)
        fit = optimization.curve_fit(f0dirshort, T, f,x0,sigma=sigma,absolute_sigma = True)
    else:
        fit = optimization.curve_fit(f0dirshort, T, f, x0)
    return fit

def fit_tc_brute(t,df_over_f,nuref,tc_range = (0.5,1.5),alpha_range = (0,1), n_grid_points=100, error=None, plot = True,Verbose = False,**keywords):
    '''
    brute force fitter for fitting Tc and alpha assuming gamma =1
    but of alpha and gamma are degenerate
    t is temperature in kelvin
    df_over_f is f-f0/f0 
    '''

    if error is None:
        error = np.ones(len(t))

    tc_values = np.linspace(tc_range[0], tc_range[1], n_grid_points)
    alpha_values = np.linspace(alpha_range[0], alpha_range[1], n_grid_points)
    evaluated_ranges = np.vstack((tc_values, alpha_values))

    a, b= np.meshgrid(tc_values, alpha_values, indexing="ij")  # always index ij

    #evaluated = noise_profile_lor_vec(x[index_for_fitting], a, b, c, d)
    evaluated = deltaf_f_vec(t,a,nuref,b,1)
    data_values = np.reshape(df_over_f, (df_over_f.shape[0], 1, 1))
    error = np.reshape(error, (error.shape[0], 1, 1))
    #print(evaluated.shape)
    # print(data_values.shape)
    # print(error.shape)
    sum_dev = np.sum(((evaluated - data_values) ** 2 / error ** 2),
                     axis=0)  # comparing in magnitude space rather than magnitude squared


    min_index = np.where(sum_dev == np.min(sum_dev))
    if Verbose:
        print("grid values at minimum are")
        print(min_index)
    index1 = min_index[0][0]
    index2 = min_index[1][0]
    fit_values = np.asarray((tc_values[index1], alpha_values[index2]))

    fit_values_names = ('tc', 'alpha')
    fit_result = deltaf_f(t, tc_values[index1], nuref,alpha_values[index2],1)

    if plot:
        extent = [evaluated_ranges[1,0],evaluated_ranges[1,n_grid_points-1],evaluated_ranges[0,0],evaluated_ranges[0,n_grid_points-1]]
        aspect = (evaluated_ranges[1,n_grid_points-1]-evaluated_ranges[1,0])/(evaluated_ranges[0,n_grid_points-1]-evaluated_ranges[0,0])
        plt.figure()
        plt.imshow(np.log10(sum_dev-np.min(sum_dev)),extent =extent,aspect = aspect,origin = 'lower', cmap = 'jet')
        plt.xlabel("alpha")
        plt.ylabel("Tc")
        plt.colorbar(label = 'Log10(sum residuals squared)')

    fit_dict = {'fit_values': fit_values, 'fit_values_names': fit_values_names, 'sum_dev': sum_dev,
                'fit_result': fit_result,'evaluated_ranges': evaluated_ranges}  #'marginalized_2d':marginalized_2d,'marginalized_1d':marginalized_1d,
    return fit_dict



import numpy as np
from scipy.stats import binned_statistic
import scipy.optimize as optimization
import matplotlib.pyplot as plt

#set of modules for fitting psd of kinetic inductance detectors
#Written by Jordan 1/5/2017

#To Do
#add verbose = true keyword

#Change Log
#1/9/2017 Added sigma_increase_cutoff and sigma_increase_factor to fit_psd modules


# noise profiles
def noise_profile(y,a,b,c):
    return a+b*y**-c
def noise_slope(y,b,c):
    return b*y**-c
def noise_white(y,a):
    return np.ones(y.shape[0])*a
def noise_profile_lor(y,a,b,c,d):
    return (a+b*y**-c)/(1+(2*np.pi*y*d)**2.)
def noise_profile_lor_vec(y,a,b,c,d):
    y = np.reshape(y,(y.shape[0],1,1,1,1))
    return (a+b*y**-c)/(1+(2*np.pi*y*d)**2.)

def std_of_mean(x):
    if np.size(x) == 1:
        return x
    else:
        return np.std(x)/np.sqrt(np.size(x))

    index_for_fitting = np.where(((x>=freq_range[0]) & (x <=freq_range[1])))
        
    if ('ranges' in keywords):
        ranges = keywords['ranges']
    else:
        if ('white_freq' in keywords):
            white_freq = keywords['white_freq']
            white_index = np.argmin(np.abs(x[index_for_fitting]-white_freq))
        else:
            white_index = len(y[index_for_fitting])//2 # noise is white in the middle of psd

        white_guess = y[index_for_fitting][white_index]
        x0_guess = np.array([ white_guess,
                                 (y[index_for_fitting][0]-white_guess)/x[index_for_fitting][0]**(-1.), # assume 1/f dominates at lowest frequency but still subtract off white noise
                                 1,# guess 1/f is index is 1
                                 1./2/np.pi/x[index_for_fitting][np.argmin(np.abs(white_guess/2.-y[index_for_fitting]))]]) # look for 3dB decrease from white noise guess
        print("guess values are")
        print(x0_guess)
        ranges = np.asarray(([x0_guess[0]/2,x0_guess[1]/20,0.5,x0_guess[3]/4],[ x0_guess[0]*2.0,x0_guess[1]*10,2,x0_guess[3]*4]))
                

    if error is None:
        error = np.ones(len(x[index_for_fitting]))

    a_values = np.linspace(ranges[0][0],ranges[1][0],n_grid_points)
    b_values = np.linspace(ranges[0][1],ranges[1][1],n_grid_points)
    c_values = np.linspace(ranges[0][2],ranges[1][2],n_grid_points)
    d_values = np.linspace(ranges[0][3],ranges[1][3],n_grid_points)
    evaluated_ranges = np.vstack((a_values,b_values,c_values,d_values))

    a,b,c,d = np.meshgrid(a_values,b_values,c_values,d_values,indexing = "ij") #always index ij

    evaluated = noise_profile_lor_vec(x[index_for_fitting],a,b,c,d)
    data_values = np.reshape(y[index_for_fitting],(y[index_for_fitting].shape[0],1,1,1,1))
    error = np.reshape(error,(y[index_for_fitting].shape[0],1,1,1,1))
    #print(evaluated.shape)
    #print(data_values.shape)
    #print(error.shape)
    sum_dev = np.sum(((evaluated-data_values)**2/error**2),axis = 0) # comparing in magnitude space rather than magnitude squared
    #print(sum_dev.shape)

    min_index = np.where(sum_dev == np.min(sum_dev))
    print("grid values at minimum are")
    print(min_index)
    index1 = min_index[0][0]
    index2 = min_index[1][0]
    index3 = min_index[2][0]
    index4 = min_index[3][0]
    fit_values = np.asarray((a_values[index1],b_values[index2],c_values[index3],d_values[index4]))
    fit_values_names = ('a (white)','b (1/f)','c (1/f exponent)','d (tau)')
    fit_result = noise_profile_lor(x,a_values[index1],b_values[index2],c_values[index3],d_values[index4])
    x0_guess_result = noise_profile_lor(x,x0_guess[0],x0_guess[1],x0_guess[2],x0_guess[3])
    noise_slope_result = noise_slope(x,fit_values[1],fit_values[2])
    fine_freqs = np.logspace(np.log10(freq_range[0]),np.log10(freq_range[1]),10000)
    print(fine_freqs)
    knee = fine_freqs[np.argmin(np.abs( fit_values[1]*fine_freqs**-fit_values[2]-fit_values[0] ))]

    fit_dict = {'fit_values': fit_values,'fit_values_names':fit_values_names, 'sum_dev': sum_dev, 'fit_result': fit_result,'x0_guess_result':x0_guess_result,'evaluated_ranges':evaluated_ranges,'knee':knee,'noise_slope_result':noise_slope_result}#, 'x0':x0, 'z':z},'marginalized_2d':marginalized_2d,'marginalized_1d':marginalized_1d,
    return fit_dict   

    return fit_dict

def fit_psd_lor(x,y,**keywords):
    '''
    # keywards are
    # use_range ---is an n length tuple of frequencies to use while fitting
    # Example: [[1,57],[63,117],[123,177]] here we fit from 1 to 57Hz and 63 to 117 Hz and 123 to 177Hz avoid 60 Hz and harmonics
    # x0    --- intial guess for the fit this can be very important becuase because least square space over all the parameter is comple
    # there are two way the function if fit one is it is fit without binning. In this case a error for the fit is calculated
    # by calculating the standard deviation of the surronding 50 or so points. Then the data below 10Hz has the error artificially
    # lowered so that the fitter doesn't ignore in when comparing it to the many more points at higher frequencies. The other way is to use
    # the log keyword which then log bins the data and calculates the error for each bin. This way there are around the same number of points
    # at low frequency as compared to high frequency
    # since there are less points at low frequencies than high frequencies I artificially increase the accuracy of the low frequency points
    # below sigma_increase_cutoff by scaling the error that the fitter uses by sigma_increase_factor
    '''
    if ('sigma_increase_cutoff' in keywords):
        sigma_increase_cutoff = keywords['sigma_increase_cutoff']
    else:
        #define default bounds
        sigma_increase_cutoff = 2. #(Hz)
    if ('sigma_increase_cutoff' in keywords):
        sigma_increase_factor = keywords['sigma_increase_factor']
    else:
        #define default bounds
        sigma_increase_factor = 5. 

    # bounds with out these some parameter might converge to non physical values
    if ('bounds' in keywords):
        bounds = keywords['bounds']
    else:
        #define default bounds
        print("default bounds used")
        bounds = ([10**-20,0.0,0.0,9e-7],[10**-8,10**-8,3,0.001]) 
    if ('use_range' in keywords):
        use_range = keywords['use_range']
        # create an index of the values you want to fit
        index = np.where((x>use_range[0][0]) & (x<use_range[0][1]) )[0]
        for i in range(1,len(use_range)):
            index2 = np.where((x>use_range[i][0]) & (x<use_range[i][1]) )
            index = np.hstack((index,index2[0]))  
    else:
        index = range(0,x.shape[0])

    # initial conditions    
    if ('x0' in keywords):
        x0 = keywords['x0']
    else:
        #define default intial guess
        print("default initial guess used")        
        x0  = np.array([1.*10**(-15.75), 1.*10**(-17),1,0.00001]) # default values that work OK for superspec

    # log bin the data first or no    
    if ('log' in keywords):
        print("psd will be log binned before fitting")
        log = 1
        bins = np.logspace(np.log10(x[0]),np.log10(x[x.shape[0]-1]),100) #100 logspaced bins 
    else:
        log = 0

    if log == 1:
        binnedfreq_temp =  binned_statistic(x[index], x[index], bins=bins)[0]
        binnedvals_temp = binned_statistic(x[index], y[index], bins=bins)[0]
        binnedvals_std = binned_statistic(x[index], y[index], bins=bins, statistic = std_of_mean)[0]
        binnedfreq = binnedfreq_temp[~np.isnan(binnedfreq_temp)]
        binnedvals = binnedvals_temp[~np.isnan(binnedfreq_temp)]
        binnedstd = binnedvals_std[~np.isnan(binnedfreq_temp)]

    freqs = x[index]
    vals = y[index]

    if log ==0: #when fitting there are so many points at high frequencies compared to at low frequecies a fitting will almost ingnore the low frequency end
        # I get an extimate fo the noise by taking the standard deviation of each 10 consective points (will be some error for th last 10 points)
        std_pts = 100 # if this number is to low it seems to bias the fits to the lower side
        low_freq_index = np.where(freqs<sigma_increase_cutoff)
        temp = np.zeros((vals.shape[0],std_pts))
        # here I estimate the error by looking at the 100 surronding points and calculated the std      
        for i in range(0,std_pts):
            temp[:,i] = np.roll(vals,-i)
        sigma = np.std(temp,axis = 1)
        sigma[low_freq_index] = sigma[low_freq_index]/sigma_increase_factor # artificial pretend the noise at low frequcies is 5 time lower than every where else
        fit = optimization.curve_fit(noise_profile_lor, freqs, vals, x0 , sigma,bounds = bounds)
    else:
        sigma = binnedstd
        fit = optimization.curve_fit(noise_profile_lor, binnedfreq, binnedvals, x0 ,sigma,bounds = bounds)

    return fit


def fit_psd_lor_brute(x, y, n_grid_points=20, error=None, **keywords):
    """
    brute force fitting (the only way to fit with 4 or less variables)
    x is the psd frequency
    y is the psd magnitude
    n_grid_points is the deminsion of the 4 dimensional cube over which function will be evaluated
    keywords are
    ranges is the ranges for each parameter i.e. np.asarray(([a_low,b_low,c_low,d_low],[a_high,b_high,c_high,d_high]))
    freq_range = (f_low,f_high) bounds over which the psd should be fit
    white_freq = 50 (i.e) Hz if using the automated range it is very useful to specify where the noise psd is white (frequency independant)
    use_range ---is an n length tuple of frequencies to use while fitting
    Example: [[1,57],[63,117],[123,177]] here we fit from 1 to 57Hz and 63 to 117 Hz and 123 to 177Hz avoid 60 Hz and harmonic
    slope: looking for TLS with a exponent of 0.5 the just force slope to be slope = 0.5

    due to the vector nature of the calculations used by this brute for fitter
    n_grid_points will be limited by your computers ram and it grows fast
    for example 50^4*(2bytes per float*(2arrays) = 3.5GB of ram
    just watch your resources when you fit if you exceed your ram you will write to disk and the fit will never finish

    To Do add in marginilaztions for error bars like in the brute force fitter in resonance fitting
    also add in the corner plot for marginalized values
    would be good to add in a nested version of this where it fits again over a smaller parameter space
    -Done for slope should think about making it general
    """

    if ('slope' in keywords):
        slope = keywords['slope']
    else:
        slope = None

    if ('freq_range' in keywords):
        freq_range = keywords['freq_range']
    else:
        freq_range = (x[0], x[-1])

    if ('use_range' in keywords):
        use_range = keywords['use_range']
        # create an index of the values you want to fit
        index_for_fitting = np.where((x>use_range[0][0]) & (x<use_range[0][1]) )[0]
        for i in range(1,len(use_range)):
            index2 = np.where((x>use_range[i][0]) & (x<use_range[i][1]) )
            index_for_fitting = np.hstack((index_for_fitting,index2[0]))  
    else:
        index_for_fitting = np.where(((x >= freq_range[0]) & (x <= freq_range[1])))
 

    if ('ranges' in keywords):
        ranges = keywords['ranges']
    else:
        if ('white_freq' in keywords):
            white_freq = keywords['white_freq']
            white_index = np.argmin(np.abs(x[index_for_fitting] - white_freq))
        else:
            white_index = len(y[index_for_fitting]) // 2  # noise is white in the middle of psd

        # only look at frequencies greater than white noise frequency
        index_for_time_constant = np.where(x >= x[index_for_fitting][white_index])

        white_guess = y[index_for_fitting][white_index]
        # assume 1/f dominates at lowest frequency but still subtract off white noise
        one_over_f_guess = (y[index_for_fitting][0] - white_guess) / x[index_for_fitting][0] ** (-1.)
        if one_over_f_guess<0: #unpysical to be less than 0 must be small though
            one_over_f_guess = white_guess/10.
        x0_guess = np.array([white_guess,
                             one_over_f_guess,
                             # assume 1/f dominates at lowest frequency but still subtract off white noise
                             1,  # guess 1/f is index is 1
                             1. / 2 / np.pi / x[index_for_time_constant][np.argmin(np.abs(white_guess / 2. - y[
                                 index_for_time_constant]))]])  # look for 3dB decrease from white noise guess
        print("guess values are")
        print(x0_guess)
        ranges = np.asarray(([x0_guess[0] / 2, x0_guess[1] / 20, 0.4, x0_guess[3] / 2],
                             [x0_guess[0] * 2, x0_guess[1] * 10, 2, x0_guess[3] * 2]))

    if error is None:
        error = np.ones(len(x[index_for_fitting]))
      
    if not slope:
        a_values = np.linspace(ranges[0][0], ranges[1][0], n_grid_points)
        b_values = np.linspace(ranges[0][1], ranges[1][1], n_grid_points)
        c_values = np.linspace(ranges[0][2], ranges[1][2], n_grid_points)
        d_values = np.linspace(ranges[0][3], ranges[1][3], n_grid_points)
        evaluated_ranges = np.vstack((a_values, b_values, c_values, d_values))

        a, b, c, d = np.meshgrid(a_values, b_values, c_values, d_values, indexing="ij")  # always index ij

        evaluated = noise_profile_lor_vec(x[index_for_fitting], a, b, c, d)
        
        data_values = np.reshape(y[index_for_fitting], (y[index_for_fitting].shape[0], 1, 1, 1, 1))
        error = np.reshape(error, (y[index_for_fitting].shape[0], 1, 1, 1, 1))
        print(evaluated.shape)
        print(data_values.shape)
        print(error.shape)
        sum_dev = np.sum(((evaluated - data_values) ** 2 / error ** 2),
                         axis=0)  # comparing in magnitude space rather than magnitude squared
        # print(sum_dev.shape)

        min_index = np.where(sum_dev == np.min(sum_dev))
        print("grid values at minimum are")
        print(min_index)
        index1 = min_index[0][0]
        index2 = min_index[1][0]
        index3 = min_index[2][0]
        index4 = min_index[3][0]

        fit_values = np.asarray((a_values[index1], b_values[index2], c_values[index3], d_values[index4]))
    else:
        a_values = np.linspace(ranges[0][0], ranges[1][0], n_grid_points)
        b_values = np.linspace(ranges[0][1], ranges[1][1], n_grid_points)
        #c_values = np.linspace(ranges[0][2], ranges[1][2], n_grid_points)
        d_values = np.linspace(ranges[0][3], ranges[1][3], n_grid_points)
        evaluated_ranges = np.vstack((a_values, b_values, d_values))

        a, b, d = np.meshgrid(a_values, b_values, d_values, indexing="ij")  # always index ij
      
        noise_profile_fixed_slope = lambda x_lamb,a,b,c,d: noise_profile_lor_vec(x_lamb,a,b,slope,d)
        evaluated = noise_profile_fixed_slope(x[index_for_fitting],a,b,slope,d)
        evaluated = evaluated[:,0,:,:,:]
        data_values = np.reshape(y[index_for_fitting], (y[index_for_fitting].shape[0], 1, 1, 1))
        error = np.reshape(error, (y[index_for_fitting].shape[0], 1, 1, 1))
        
        # print(evaluated.shape)
        # print(data_values.shape)
        # print(error.shape)
        
        sum_dev = np.sum(((evaluated - data_values) ** 2 / error ** 2),
                         axis=0)  # comparing in magnitude space rather than magnitude squared
        # print(sum_dev.shape)

        min_index = np.where(sum_dev == np.min(sum_dev))
        print("grid values at minimum are")
        print(min_index)
        index1 = min_index[0][0]
        index2 = min_index[1][0]
        #index3 = min_index[2][0]
        index4 = min_index[2][0]
        
        fit_values = np.asarray((a_values[index1], b_values[index2], slope, d_values[index4]))

    fit_values_names = ('a (white)', 'b (1/f)', 'c (1/f exponent)', 'd (tau)')
    fit_result = noise_profile_lor(x, fit_values[0], fit_values[1], fit_values[2], fit_values[3])
    x0_guess_result = noise_profile_lor(x, x0_guess[0], x0_guess[1], x0_guess[2], x0_guess[3])
    noise_slope_result = noise_slope(x, fit_values[1], fit_values[2])
    fine_freqs = np.logspace(np.log10(freq_range[0]), np.log10(freq_range[1]), 10000)
    #print(fine_freqs)
    knee = fine_freqs[np.argmin(np.abs(fit_values[1] * fine_freqs ** -fit_values[2] - fit_values[0]))]

    fit_dict = {'fit_values': fit_values, 'fit_values_names': fit_values_names, 'sum_dev': sum_dev,
                'fit_result': fit_result, 'x0_guess_result': x0_guess_result, 'evaluated_ranges': evaluated_ranges,
                'knee': knee,
                'noise_slope_result': noise_slope_result}  # , 'x0':x0, 'z':z},'marginalized_2d':marginalized_2d,'marginalized_1d':marginalized_1d,
    return fit_dict


def fit_psd(x,y,plot = False,**keywords):
    """
    # keywards are
    # use_range ---is an n length tuple of frequencies to use while fitting
    # Example: [[1,57],[63,117],[123,177]] here we fit from 1 to 57Hz and 63 to 117 Hz and 123 to 177Hz avoid 60 Hz and harmonics
    # x0    --- intial guess for the fit this can be very important becuase because least square space over all the parameter is comple
    # there are two way the function if fit one is it is fit without binning. In this case a error for the fit is calculated
    # by calculating the standard deviation of the surronding 50 or so points. Then the data below 10Hz has the error artificially
    # lowered so that the fitter doesn't ignore in when comparing it to the many more points at higher frequencies. The other way is to use
    # the log keyword which then log bins the data and calculates the error for each bin. This way there are around the same number of points
    # at low frequency as compared to high frequency
    # since there are less points at low frequencies than high frequencies I artificially increase the accuracy of the low frequency points
    # below sigma_increase_cutoff by scaling the error that the fitter uses by sigma_increase_factor
    """
    if ('sigma_increase_cutoff' in keywords):
        sigma_increase_cutoff = keywords['sigma_increase_cutoff']
    else:
        #define default bounds
        sigma_increase_cutoff = 2. #(Hz)
    if ('sigma_increase_cutoff' in keywords):
        sigma_increase_factor = keywords['sigma_increase_factor']
    else:
        #define default bounds
        sigma_increase_factor = 5. 

    index_for_fitting = np.where(((x>=freq_range[0]) & (x <=freq_range[1])))
        
    if ('ranges' in keywords):
        ranges = keywords['ranges']
    else:
        if ('white_freq' in keywords):
            white_freq = keywords['white_freq']
            white_index = np.argmin(np.abs(x[index_for_fitting]-white_freq))
        else:
            white_index = len(y[index_for_fitting])//2 # noise is white in the middle of psd

        white_guess = y[index_for_fitting][white_index]
        x0_guess = np.array([ white_guess,
                                 (y[index_for_fitting][0]-white_guess)/x[index_for_fitting][0]**(-1.), # assume 1/f dominates at lowest frequency but still subtract off white noise
                                 1,# guess 1/f is index is 1
                                 1./2/np.pi/x[index_for_fitting][np.argmin(np.abs(white_guess/2.-y[index_for_fitting]))]]) # look for 3dB decrease from white noise guess
        print("guess values are")
        print(x0_guess)
        ranges = np.asarray(([x0_guess[0]/2,x0_guess[1]/20,0.5,x0_guess[3]/2],[ x0_guess[0]*2,x0_guess[1]*10,2,x0_guess[3]*2]))
                

    if error is None:
        error = np.ones(len(x[index_for_fitting]))

    a_values = np.linspace(ranges[0][0],ranges[1][0],n_grid_points)
    b_values = np.linspace(ranges[0][1],ranges[1][1],n_grid_points)
    c_values = np.linspace(ranges[0][2],ranges[1][2],n_grid_points)
    d_values = np.linspace(ranges[0][3],ranges[1][3],n_grid_points)
    evaluated_ranges = np.vstack((a_values,b_values,c_values,d_values))

    a,b,c,d = np.meshgrid(a_values,b_values,c_values,d_values,indexing = "ij") #always index ij

    evaluated = noise_profile_lor_vec(x[index_for_fitting],a,b,c,d)
    data_values = np.reshape(y[index_for_fitting],(y[index_for_fitting].shape[0],1,1,1,1))
    error = np.reshape(error,(y[index_for_fitting].shape[0],1,1,1,1))
    #print(evaluated.shape)
    #print(data_values.shape)
    #print(error.shape)
    sum_dev = np.sum(((evaluated-data_values)**2/error**2),axis = 0) # comparing in magnitude space rather than magnitude squared
    #print(sum_dev.shape)

    min_index = np.where(sum_dev == np.min(sum_dev))
    print("grid values at minimum are")
    print(min_index)
    index1 = min_index[0][0]
    index2 = min_index[1][0]
    index3 = min_index[2][0]
    index4 = min_index[3][0]
    fit_values = np.asarray((a_values[index1],b_values[index2],c_values[index3],d_values[index4]))
    fit_values_names = ('a (white)','b (1/f)','c (1/f exponent)','d (tau)')
    fit_result = noise_profile_lor(x,a_values[index1],b_values[index2],c_values[index3],d_values[index4])
    x0_guess_result = noise_profile_lor(x,x0_guess[0],x0_guess[1],x0_guess[2],x0_guess[3])
    noise_slope_result = noise_slope(x,fit_values[1],fit_values[2])
    fine_freqs = np.logspace(np.log10(freq_range[0]),np.log10(freq_range[1]),10000)
    print(fine_freqs)
    knee = fine_freqs[np.argmin(np.abs( fit_values[1]*fine_freqs**-fit_values[2]-fit_values[0] ))]

    fit_dict = {'fit_values': fit_values,'fit_values_names':fit_values_names, 'sum_dev': sum_dev, 'fit_result': fit_result,'x0_guess_result':x0_guess_result,'evaluated_ranges':evaluated_ranges,'knee':knee,'noise_slope_result':noise_slope_result}#, 'x0':x0, 'z':z},'marginalized_2d':marginalized_2d,'marginalized_1d':marginalized_1d,
    return fit_dict   

    return fit_dict

    if ('use_range' in keywords):
        use_range = keywords['use_range']
        # create an index of the values you want to fit
        index = np.where((x>use_range[0][0]) & (x<use_range[0][1]) )[0]
        for i in range(1,len(use_range)):
            index2 = np.where((x>use_range[i][0]) & (x<use_range[i][1]) )
            index = np.hstack((index,index2[0]))  
    else:
        index = range(0,x.shape[0])

    if ('uniform_weight' in keywords):
        uniform_weight = keywords['uniform_weight']
    else:
        uniform_weight = False

    # initial conditions    
    if ('x0' in keywords):
        x0 = keywords['x0']
    else:
        #define default intial guess # guess for b: b*freqs**-c = value => b = value/freqs**-c
        print("default initial guess used")        
        x0  = np.array([y[index][-1], y[index][0]/x[index][0]**(-1.) ,1]) # default values that work OK for superspec
    # bounds with out these some parameter might converge to non physical values
    if ('bounds' in keywords):
        bounds = keywords['bounds']
    else:
        #define default bounds
        print("default bounds used")
        #bounds = ([10**-20,10**-20,0],[10**-10,10**-10,3])
        bounds = ([x0[0]/10.,x0[1]/10.,0],[x0[0]*10,x0[1]*10.,3])

    # log bin the data first or no    
    if ('log' in keywords):
        print("psd will be log binned before fitting")
        log = 1
        bins = np.logspace(np.log10(x[0]),np.log10(x[x.shape[0]-1]),100) #100 logspaced bins 
    else:
        log = 0
        bins = x.shape[0] # doesn't bin at all

 
    binnedfreq_temp =  binned_statistic(x[index], x[index], bins=bins)[0]
    binnedvals_temp = binned_statistic(x[index], y[index], bins=bins)[0]
    binnedvals_std = binned_statistic(x[index], y[index], bins=bins, statistic = std_of_mean)[0]
    binnedfreq = binnedfreq_temp[~np.isnan(binnedfreq_temp)]
    binnedvals = binnedvals_temp[~np.isnan(binnedfreq_temp)]
    binnedstd = binnedvals_std[~np.isnan(binnedfreq_temp)]

    freqs = x[index]
    vals = y[index] 

    if log ==0: #when fitting there are so many points at high frequencies compared to at low frequecies a fitting will almost ingnore the low frequency end
        # I get an extimate fo the noise by taking the standard deviation of each 10 consective points (will be some error for th last 10 points)
        std_pts = 100 # if this number is to low it seems to bias the fits to the lower side
        low_freq_index = np.where(freqs<sigma_increase_cutoff)
        temp = np.zeros((vals.shape[0],std_pts))       
        for i in range(0,std_pts):
            temp[:,i] = np.roll(vals,-i)
        sigma = np.std(temp,axis = 1)
        sigma[low_freq_index] = sigma[low_freq_index]/sigma_increase_factor # artificial pretend the noise at low frequcies is 10 time lower than every where else
        fit = optimization.curve_fit(noise_profile, freqs, vals, x0 , sigma,bounds = bounds)
        print("hello")
    else:
        if uniform_weight == True:
            sigma = np.ones(len(binnedstd))*np.mean(binnedstd)

        else:
            sigma = binnedstd
        fit = optimization.curve_fit(noise_profile, binnedfreq, binnedvals, x0 ,sigma,bounds = bounds)

    if plot == True:
        plt.loglog(binnedfreq,binnedvals,label = "Data",linewidth = 2)
        #plt.errorbar(binnedfreq,binned_psd_log,binned_std_log, fmt='o')
        plt.loglog(binnedfreq,noise_profile(binnedfreq,x0[0],x0[1],x0[2]),linewidth = 2,label = "Initial Guess")
        plt.loglog(binnedfreq,noise_profile(binnedfreq,fit[0][0],fit[0][1],fit[0][2]),linewidth = 2,label = "Fit")
        plt.loglog(binnedfreq,noise_slope(binnedfreq,fit[0][1],fit[0][2]),linewidth = 2, label = "1/f^" + str(fit[0][2])[0:4])
        plt.loglog(binnedfreq,noise_white(binnedfreq,fit[0][0]),linewidth = 2,label = "White"+ " " +  str(fit[0][0]*10**16)[0:4] +" x10^-16")
        plt.legend()
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Sxx (1/Hz)")
        plt.ylim(np.min(y),np.max(y))
        plt.show(block = False)
        

    return fit




from submm.lab_brick import core

#to use this do 
#sudo python
#execfile("labbrick_startup.py")

#attn2 = core.Attenuator(0x041f,0x1208,"16776")
#attn1 = core.Attenuator(0x041f,0x1208,"15915")
attn2 = core.Attenuator(0x041f, 0x1208, "01784")


print("15915 is attn1")
print("16776 is attn2")
print("example command: attn1.get_attenuation()")
print("example command: attn2.set_attenuation(29)")



import numpy as np
from scipy.stats import binned_statistic
import matplotlib.pyplot as plt
from scipy.io.idl import readsav as readsav
from submm.KIDs import psd_fitting

# this is module to demostrate how to used the psd fitting code
#written by Jordan Wheeler on 1/6/2017

#Change Log


bins = 10000
log_bins = np.logspace(0,4,100) 

# get some data to test the fitter on (this is single tone data processed by Steve Hailey-Dunsheath's idl code)
raw = readsav("./20151106_noise_07/Set0000Pn00Fn00/rawdata.sav",verbose = False)
data = readsav("./20151106_noise_07/Set0000Pn00Fn00/psddata.sav",verbose = False)

freqs_raw = data['mixer_data_psd']['s_xx'][0]['freq_vec'][0]
psd_raw = data['mixer_data_psd']['s_xx'][0]['csd_xy'][0]

#throw out negative frequecies
greater_than_zero = np.where(freqs_raw>0)
freqs = freqs_raw[greater_than_zero]
vals = np.abs(psd_raw[greater_than_zero])

#do some binning for plotting
binned_freq =  binned_statistic(freqs, freqs, bins=bins)[0]
binned_psd = binned_statistic(freqs, vals, bins=bins)[0]

#this binning is like the binning used when log = True for for psd_fitting
binned_freq_log =  binned_statistic(freqs, freqs, bins=log_bins)[0]
binned_psd_log = binned_statistic(freqs, vals, bins=log_bins)[0]
binned_std_log = binned_statistic(freqs, vals, bins=log_bins, statistic = psd_fitting.std_of_mean)[0]

mean_index = np.where((freqs>200) & (freqs<1000))
white_guess = np.mean(vals[mean_index])

# since we are not doing a facny fit the intial guess is important or else may get stuck in a local min
x0    = np.array([1.*10**(-15.75), white_guess,1,0.00001]) #intitial guess not required will use default if not specified
use_range = [[1,10],[25,4000]] # don't want to use the data from 10 to 25Hz as there is a big ugly noise spike there
# and I don't want to use the data that is to high frequency because it seems to confuse the fitter

# fitting using unbinned data
fit = psd_fitting.fit_psd_lor(freqs, vals, x0 = x0, use_range = use_range)
print(fit[0])

plt.figure(1,figsize = (12,12))
plt.subplot(221)

plt.loglog(binned_freq,binned_psd,label = "Data",linewidth = 2)
plt.loglog(freqs, psd_fitting.noise_profile_lor(freqs, fit[0][0], fit[0][1], fit[0][2], fit[0][3]), linewidth = 2, label ="Fit")
plt.loglog(freqs, psd_fitting.noise_slope(freqs, fit[0][1], fit[0][2]), linewidth = 2, label ="1/f^" + str(fit[0][2])[0:4])
plt.loglog(freqs, psd_fitting.noise_white(freqs, fit[0][0]), linewidth = 2, label ="White" + " " + str(fit[0][0] * 10 ** 16)[0:4] + " x10^-16")

plt.title("psd fit")    
plt.legend(loc = 1)
plt.xlabel("Frequency (Hz)")
plt.ylabel("Noise Sxx (1/Hz)")
plt.xlim(1,10**4)
plt.ylim(10**-17,10**-13)

#fitting after log binning the data
fit = psd_fitting.fit_psd_lor(freqs, vals, x0 = x0, use_range = use_range, log = True, uniform_weight = True)
print(fit[0])

plt.subplot(222)

plt.loglog(binned_freq,binned_psd,label = "Data",linewidth = 2)
plt.errorbar(binned_freq_log,binned_psd_log,binned_std_log, fmt='o')
plt.loglog(freqs, psd_fitting.noise_profile_lor(freqs, fit[0][0], fit[0][1], fit[0][2], fit[0][3]), linewidth = 2, label ="Fit")
plt.loglog(freqs, psd_fitting.noise_slope(freqs, fit[0][1], fit[0][2]), linewidth = 2, label ="1/f^" + str(fit[0][2])[0:4])
plt.loglog(freqs, psd_fitting.noise_white(freqs, fit[0][0]), linewidth = 2, label ="White" + " " + str(fit[0][0] * 10 ** 16)[0:4] + " x10^-16")

plt.title("psd fit log binned before fitting")
plt.xlabel("Frequency (Hz)")
plt.ylabel("Noise Sxx (1/Hz)")  
plt.legend(loc = 1)
plt.xlim(1,10**4)
plt.ylim(10**-17,10**-13)


#fitting without lorentzian fall off
use_range = [[1,10],[25,500]]
x0    = np.array([1.*10**(-15.75), white_guess,1]) #intitial guess not required will use default if not specified

fit = psd_fitting.fit_psd(freqs, vals, x0 = x0, use_range = use_range)
print(fit[0])

plt.subplot(223)

plt.loglog(binned_freq,binned_psd,label = "Data",linewidth = 2)
#plt.errorbar(binned_freq_log,binned_psd_log,binned_std_log, fmt='o')
plt.loglog(freqs, psd_fitting.noise_profile(freqs, fit[0][0], fit[0][1], fit[0][2]), linewidth = 2, label ="Fit")
plt.loglog(freqs, psd_fitting.noise_slope(freqs, fit[0][1], fit[0][2]), linewidth = 2, label ="1/f^" + str(fit[0][2])[0:4])
plt.loglog(freqs, psd_fitting.noise_white(freqs, fit[0][0]), linewidth = 2, label ="White" + " " + str(fit[0][0] * 10 ** 16)[0:4] + " x10^-16")

plt.title("psd fit without lorentzian")
plt.xlabel("Frequency (Hz)")
plt.ylabel("Noise Sxx (1/Hz)")  
plt.legend(loc = 1)
plt.xlim(1,10**4)
plt.ylim(10**-17,10**-13)

#fitting without lorentzian fall off and log binning before fit
fit = psd_fitting.fit_psd(freqs, vals, x0 = x0, use_range = use_range, log = True)
print(fit[0])

plt.subplot(224)

plt.loglog(binned_freq,binned_psd,label = "Data",linewidth = 2)
plt.errorbar(binned_freq_log,binned_psd_log,binned_std_log, fmt='o')
plt.loglog(freqs, psd_fitting.noise_profile(freqs, fit[0][0], fit[0][1], fit[0][2]), linewidth = 2, label ="Fit")
plt.loglog(freqs, psd_fitting.noise_slope(freqs, fit[0][1], fit[0][2]), linewidth = 2, label ="1/f^" + str(fit[0][2])[0:4])
plt.loglog(freqs, psd_fitting.noise_white(freqs, fit[0][0]), linewidth = 2, label ="White" + " " + str(fit[0][0] * 10 ** 16)[0:4] + " x10^-16")

plt.title("psd fit without lorentzian")
plt.xlabel("Frequency (Hz)")
plt.ylabel("Noise Sxx (1/Hz)")  
plt.legend(loc = 1)
plt.xlim(1,10**4)
plt.ylim(10**-17,10**-13)

plt.show()



import usb.core #pip install pyusb
import threading
from datetime import datetime
import sys
import numpy as np
import platform


'''
Got this of some guys github and made a few edits 
see https://github.com/markafarrell/lab-brick-attenuator.git for the original
-edited so that it can connect to more that one labbrick that have different serial numbers
-edited to put the call back function in the core file for attenuation_change and attenuation response
-edited to return the value for get_attenuation()

Currently all other function besides get_attenuation and set_attenuation will fail unless 
User deffined callback are defined for each commmand or that part of the command is commented 
out example(set_attenuation is currently commmented out)
'''

def attenuation_change(device, status_byte, count, byteblock, val):
	print("User Defined Callback for attenuation_level_change")
	print("Attenuation: " + str(val))
	
def attenuation_response(device, status_byte, count, byteblock, val):
	#print("User Defined Callback for attenuation_level_response")
	print("Attenuation: " + str(val))
	#return val

class Attenuator(object):
	"""Control a Vaunix Lab-brick attenuator"""

	#################################################################################
	#								USB VARIABLES									#
	#################################################################################
	
	__vid = 0x041f #0x00 
	__pid = 0x1208 #0x00
	serial = "15915" #
	
	__dev = False
	__write_endpoint = True 
	__read_endpoint = None
	
	#################################################################################
	#									VARIABLES									#
	#################################################################################
	
	attenuation_level = -1
	ramp_dwell_time = -1
	ramp_dwell_time_bidirectional = -1
	ramp_start_attenuation_level = -1
	ramp_stop_attenuation_level = -1
	ramp_step_size = -1
	ramp_step_size_bidirectional = -1
	
	ramp_status = -1
	
	# 0 - Ramping up
	# 1 - Ramping down
	
	ramp_mode = -1
	
	# 1 - Single ramp
	# 2 - Continuous ramp
	
	ramp_wait_time = -1
	ramp_wait_time_bidirectional = -1
	
	maximum_attenuation = -1
	
	save_required = -1
	
	pll_lock = -1
	
	rf_hardware_state = -1
	
	connected = False

	__read_thread = False
	
	
	#################################################################################
	#									LOGGING										#
	#################################################################################
	
	__log = sys.stderr
	__write_log_string = ">>> %s - %s \t cmd: 0x%02x \t count: %d \t byteblock: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n"
	__read_log_string = "<<< %s - %s \t res: 0x%02x \t count: %d \t byteblock: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n"
	__cmd_log_string = ">>> %s - %s \t %s\n"
	__res_log_string = "<<< %s - %s \t %s %d\n"
	__status_log_string = "<<< %s - %s \t status: 0x%02x \t dev_status: 0b%x \t attenuation: %d\n"
	
	#################################################################################
	#									CONSTRUCTOR									#
	#################################################################################
		
	def __init__(self, vid, pid,serial_number, debug=False):
	
		self.__vid = vid
		self.__pid = pid

		if debug:
			self.__log.write("vid=0x%04x pid=0x%04x\n" % (self.__vid, self.__pid)) 
	
		# Jordan edited to handle more than one labbrick
		devices = usb.core.find(idVendor=self.__vid, idProduct=self.__pid, find_all = True)


		#loop through devices to find the right serial number
		for d in devices:
			self.__dev = d
			if self.__dev is None:
				raise ValueError('Device not found')
			
			# Does not work on windows
			if platform.system() != 'Windows':
				if self.__dev.is_kernel_driver_active(0):
					try:
						self.__dev.detach_kernel_driver(0)
						print("kernel driver detached")
					except usb.core.USBError as e:
						sys.exit("Could not detach kernel driver: %s" % str(e))
			
			try:
				self.__dev.set_configuration()

				# get an endpoint instance
				cfg = self.__dev.get_active_configuration()
				intf = cfg[(0,0)]
				self.__read_endpoint = usb.util.find_descriptor(
					intf,
				# match the first OUT endpoint
					custom_match = \
					lambda e: \
						usb.util.endpoint_direction(e.bEndpointAddress) == \
						usb.util.ENDPOINT_IN)

			#Store the serial number of the lab brick
			
				self.serial = self.get_serial_number()
				print(self.serial)

			
			#Spawn read thread to handle reading for lab brick
			
				self.connected = True
			except:
				# Something bad happened during connection
				self.connected = False
			print(self.__read_endpoint)	
			#assert self.__read_endpoint is not None #comment this out was breaking code if attenuators were called in wrong order
			if self.serial == "SN:"+str(serial_number):
				print("correct device")
				break
			else:
				self.connected = False
		

		self.spawn_read_thread(debug)

		assert self.connected == True

	def spawn_read_thread(self, debug=False):
		"""	Spawn the thread to get responses from the lab brick """
		
		self.__read_thread = threading.Thread(target=self.read_data, args=(debug,))
		self.__read_thread.daemon = True
		self.__read_thread.start()
	
	def get_serial_number(self):
		"""	Get the serial number of the connected device """
		return self.__dev.serial_number
		
	#################################################################################
	#										READ									#
	#################################################################################
	
	def read_data(self, debug=False):
	
		data = None

		while True:
			try:
				data = self.__read_endpoint.read(8)
								
				status_byte = data[0]
		
				count = data[1]
				
				byteblock = bytearray( [ data[2], data[3], data[4], data[5], data[6], data[7] ] )
				
				if debug:
					self.__log.write(self.__read_log_string % ( datetime.now(), 'read_data', status_byte, count, byteblock[0], 
						byteblock[1], byteblock[2], byteblock[3], byteblock[4], byteblock[5]))
						
				self.handle_response(status_byte, count, byteblock, debug)

			except usb.core.USBError as e:
				# USB Error
				
				# TODO: Handle this
				pass
	
	#################################################################################
	#									COMMANDS									#
	#################################################################################
	
	# All commands send to the device are USB Control Commands.
	
	# self.__dev.ctrl_transfer(bmRequestType, bmRequest, wValue, wIndex, data) 
	
	""" 
		bmRequestType 	- 	0x21
		bmRequest 		-	9
		wValue			-	0x0200
		wIndex			-	0x0000
	"""
	
	
	def send_command(self, command, count, byteblock, debug=False):
	
		bytes = bytearray( [ command, count ] )

		bytes += byteblock
		
		if debug:
			self.__log.write(self.__write_log_string % ( datetime.now(), 'send_command', command, count, byteblock[0], 
				byteblock[1], byteblock[2], byteblock[3], byteblock[4], byteblock[5]))
		
		#self.__endpoint.write(bytes)
		self.__dev.ctrl_transfer( 0x21, 0x9, 0x0200, 0x0000, bytes)
		
	def set_attenuation(self, attenuation, debug=False):
		#self.register_callback('attenuation_level_change', attenuation_change) # don't need to hear back every time
	
		"""	Set attenuation level in 0.25 dB steps. 
			0x00 is the maximum power, 0x02 is 0.5dB of attenuation. 
			Note that the resoulution of the power output setting is only 0.5dB, 
			the least significant bit of the value is ignored. 
					
			Input in dB
		"""
	
		command = 0x8D
		count = 1
		
		#Round down attenuation to nearest 0.25dB interval
		val = int(attenuation * 4)
		
		byteblock = bytearray( [ val & 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_attenuation', 'attenuation=' + str(attenuation)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_attenuation(self, debug=False):
		val = self.register_callback('attenuation_level_response', attenuation_response)
		
	
		"""	Get attenuation level in 0.25 dB steps. 
			0x00 is the maximum power, 0x02 is 0.5dB of attenuation. 
			Note that the resoulution of the power output setting is only 0.5dB, 
			the least significant bit of the value is ignored. 
		"""
	
		command = 0x0D
		
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_attenuation', ""))
			
		self.send_command(command, count, byteblock, debug)
		return float(self.attenuation_level)
	
	def set_ramp_dwell_time(self, time, debug=False):
	
		"""	Set time to dwell at each attenuation level during a ramp in 1 
			millisecond intervals
		
			Input in milliseconds
		"""
	
		command = 0xB3
		count = 4
		byteblock = bytearray( [ time & 0xFF, (time >> 8) & 0xFF, (time >> 16) & 0xFF, (time >> 24) & 0xFF, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_dwell_time', 'dwell_time=' + str(time)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_dwell_time(self, debug=False):
	
		"""	Get time to dwell at each attenuation level during a ramp in 1 
			millisecond intervals
		"""
	
		command = 0x33
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_ramp_dwell_time', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def set_ramp_dwell_time_bidirectional(self, time, debug=False):
	
		"""	Set time to dwell at each attenuation level during a ramp in 1 
			millisecond intervals
		
			Input in milliseconds
		"""
	
		command = 0xB7
		count = 4
		byteblock = bytearray( [ time & 0xFF, (time >> 8) & 0xFF, (time >> 16) & 0xFF, (time >> 24) & 0xFF, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_dwell_time_bidirectional', 'dwell_time=' + str(time)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_dwell_time_bidirectional(self, debug=False):
	
		"""	Get time to dwell at each attenuation level during a ramp in 1 
			millisecond intervals
		"""
	
		#TODO: Check this command byte
		command = 0x37
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_ramp_dwell_time_bidirectional', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def set_ramp_start_attenuation_level(self, attenuation, debug=False):
	
		"""	Set start level of the ramp in 0.25 dB units
	
			Input in dB
		"""
	
		command = 0xB0
		count = 4
		
		#Round down attenuation to nearest 0.25dB interval
		val = int(attenuation * 4)
		
		byteblock = bytearray( [ val & 0xFF, (val >> 8) & 0xFF, (val >> 16) & 0xFF, (val >> 24) & 0xFF, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_start_attenuation_level', 'attenuation=' + str(attenuation)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_start_attenuation_level(self, debug=False):
	
		"""	Get start level of the ramp in 0.25 dB units """
	
		command = 0x30
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_ramp_start_attenuation_level', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def set_ramp_stop_attenuation_level(self, attenuation, debug=False):
	
		"""	Set end level of the ramp in 0.25 dB units
	
			Input in dB
		"""
	
		command = 0xB1
		count = 4
		
		#Round down attenuation to nearest 0.25dB interval
		val = int(attenuation * 4)
		
		byteblock = bytearray( [ val & 0xFF, (val >> 8) & 0xFF, (val >> 16) & 0xFF, (val >> 24) & 0xFF, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_stop_attenuation_level', 'attenuation=' + str(attenuation)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_stop_attenuation_level(self, debug=False):
	
		"""	Get end level of the ramp in 0.25 dB units """
	
		command = 0x31
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )

		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_stop_attenuation_level', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def set_ramp_step_size(self, step_size, debug=False):
	
		""" Set ramp step size in 0.25 dB units
	
			Input in dB
		"""
	
		command = 0xB2
		count = 4
		
		#Round down step size to nearest 0.25dB interval
		val = int(step_size * 4)
		
		byteblock = bytearray( [ val & 0xFF, (val >> 8) & 0xFF, (val >> 16) & 0xFF, (val >> 24) & 0xFF, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_step_size', 'step_size=' + str(step_size)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_step_size(self, step_size, debug=False):
	
		""" Get ramp step size in 0.25 dB units """
	
		command = 0x32
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_ramp_step_size', 'step_size=' + str(step_size)))
		
		self.send_command(command, count, byteblock, debug)
	
	def start_ramp(self, type, debug=False):
	
		""" Start the Ramp operation. 
			Type 
			 - 01 for a single ramp
			 - 02 for continuous ramps.
			 - 03 for continuous up then down.
		"""
	
		command = 0x89
		count = 3
		byteblock = bytearray( [ type, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'start_ramp', 'type=' + str(type)))
		
		self.send_command(command, count, byteblock, debug)
	
	def start_ramp_single(self, debug=False):
	
		""" Start single Ramp operation. """
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'start_ramp_single', ''))
			
		self.start_ramp(0x01, debug)	

	def start_ramp_continuous(self, debug=False):
	
		""" Start continuous Ramp operation. """
	
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'start_ramp_continuous', ''))
	
		self.start_ramp(0x02, debug)

	def start_ramp_down(self, debug=False):
	
		""" Start continuous Ramp operation. """
	
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'start_ramp_continuous', ''))
	
		self.start_ramp(0x05, debug)

	def start_ramp_continuous_bidirectional(self, debug=False):
	
		""" Start continuous Ramp operation. """
	
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'start_ramp_continuous', ''))
	
		self.start_ramp(0x13, debug)
	def stop_ramp(self, debug=False):
	
		""" Stop ramp operation. """ 
	
		command = 0x39
		count = 1
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'stop_ramp', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def set_ramp_wait_time(self, wait_time, debug=False):
	
		""" Set time to wait before starting a new ramp in 1 millisecond intervals, 
			when continuous ramps are in operation.
		"""
	
		command = 0xB6
		
		# Count is documented as 1 but input is specified as a DWORD count=4 makes more sense
		
		count = 4
		byteblock = bytearray( [ wait_time & 0xFF, (wait_time >> 8) & 0xFF, (wait_time >> 16) & 0xFF, (wait_time >> 24) & 0xFF, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_wait_time', 'wait_time=' + str(wait_time)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_wait_time(self, debug=False):
	
		""" Get time to wait before starting a new ramp in 1 millisecond intervals, 
			when continuous ramps are in operation.
		"""
	
		command = 0x36
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_ramp_wait_time', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def set_ramp_wait_time_bidirectional(self, wait_time, debug=False):
	
		""" Set time to wait before starting a new ramp in 1 millisecond intervals, 
			when continuous ramps are in operation.
		"""
	
		command = 0xB9
		
		# Count is documented as 1 but input is specified as a DWORD count=4 makes more sense
		
		count = 4
		byteblock = bytearray( [ wait_time & 0xFF, (wait_time >> 8) & 0xFF, (wait_time >> 16) & 0xFF, (wait_time >> 24) & 0xFF, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_wait_time_bidirectional', 'wait_time=' + str(wait_time)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_wait_time_bidirectional(self, debug=False):
	
		""" Get time to wait before starting a new ramp in 1 millisecond intervals, 
			when continuous ramps are in operation.
		"""
	
		command = 0x39
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_ramp_wait_time_bidirectional', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def set_ramp_step_size_bidirectional(self, step_size, debug=False):
	
		""" Set the step size for when stepping down.
		"""
	
		command = 0xB8
		
		#Round down step size to nearest 0.25dB interval
		val = int(step_size * 4)
		
		byteblock = bytearray( [ val & 0xFF, (val >> 8) & 0xFF, (val >> 16) & 0xFF, (val >> 24) & 0xFF, 0x00, 0x00 ] )
		
		count = 4
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'set_ramp_step_size_bidirectional', 'step_size=' + str(step_size)))
		
		self.send_command(command, count, byteblock, debug)
	
	def get_ramp_step_size_bidirectional(self, debug=False):
	
		""" Get the step size for when stepping down
		"""
	
		command = 0x38
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_ramp_step_size_bidirectional', ''))
		
		self.send_command(command, count, byteblock, debug)
	
	def restore_defaults(self, debug=False):
	
		"""	Resets all of the parameters to their factory default settings """
		
		command = 0x8F
		
		# Not sure what count or byteblock should be here as it isn't documented very well
		count = 1 
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'restore_defaults', ''))
		
		self.send_command(command, count, byteblock, debug)
		
	def get_maximum_attenuation(self, debug=False):
	
		""" Get the maximum possible attenuation value in 0.25dB units """
	
		### TODO: Confirm actual units
	
		command = 0x35
		count = 0
		byteblock = bytearray( [ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'get_maximum_attenuation', ''))
		
		self.send_command(command, count, byteblock, debug)
		
	def save_user_parameters(self, debug=False):
	
		"""	The first three bytes of the byteblock must be set to 0x42, 0x55, 0x31 
			as a key to enable the save operation. 
			Save User Parameters records the attenuation and ramp settings into 
			non-volatile memory in the Lba Brick, The Lab Brick will reload these 
			parameters when it is powered on. 
		"""
	
		command = 0x8C
		count = 3
		byteblock = bytearray( [ 0x42, 0x55, 0x31, 0x00, 0x00, 0x00 ] )
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'save_user_parameters', ''))
		
		self.send_command(command, count, byteblock, debug)
			
	#################################################################################
	#									RESPONSES									#
	#################################################################################
	
	def handle_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Handle responses for the lab brick and call the correct handler as per 
			the status bit.
		"""
	
		try:
			(self.__responses[status_byte])(self, status_byte, count, byteblock, debug)
		except KeyError:
			1+1
			#self.__log.write(self.__read_log_string % ( datetime.now(), 'read_data', status_byte, count, byteblock[0], 
#						byteblock[1], byteblock[2], byteblock[3], byteblock[4], byteblock[5]))
			#self.__log.write("Status Byte of response not recognised\n")
	
	def attenuation_level_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x04 or 0x0D
		
			Byte = attenuation level in 0.25 dB steps. 0x00 is the maximum power, 
			0x02 is 0.5dB of attenuation. Note that the resoulution of the power 
			output setting is only 0.5dB, the least significant bit of the value 
			is ignored.
		"""
	
		if status_byte == 0x0D:
			val = byteblock[0]
		elif status_byte == 0x04:
			val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		val = val/4.0

		if self.__user_defined_callbacks['attenuation_level_response'] != None:
			self.__user_defined_callbacks['attenuation_level_response'](self, status_byte, count, byteblock, val)
		
		self.attenuation_level = val
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'attenuation_level_response', 'attenuation=' + str()))
	
	def ramp_dwell_time_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x33
		
			DWORD = Time to dwell at each attenuation level during a ramp in 1 
			millisecond intervals
		"""
		
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		
		if self.__user_defined_callbacks['ramp_dwell_time_response'] != None:
			self.__user_defined_callbacks['ramp_dwell_time_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_dwell_time = val
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_dwell_time_response', 'dwell_time=' + str(self.ramp_dwell_time)))
	
	def ramp_dwell_time_bidirectional_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x37
		
			DWORD = Time to dwell at each attenuation level during a ramp in 1 
			millisecond intervals
		"""
		
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		
		if self.__user_defined_callbacks['ramp_dwell_time_bidirectional_response'] != None:
			self.__user_defined_callbacks['ramp_dwell_time_bidirectional_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_dwell_time_bidirectional = val
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_dwell_time_response', 'dwell_time=' + str(self.ramp_dwell_time)))
	
	def ramp_start_attenuation_level_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x30
		
			DWORD = Upper level of the ram in 0.25 dB units
		"""
		
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		val = val/4.0

		if self.__user_defined_callbacks['ramp_start_attenuation_level_response'] != None:
			self.__user_defined_callbacks['ramp_start_attenuation_level_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_start_attenuation_level = val/4.0
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_start_attenuation_level_response', 'attenuation=' + str(self.ramp_start_attenuation_level)))
	
	def ramp_stop_attenuation_level_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x33
		
			DWORD = Upper level of the ram in 0.25 dB units
		"""
		
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		val = val/4.0
		
		if self.__user_defined_callbacks['ramp_stop_attenuation_level_response'] != None:
			self.__user_defined_callbacks['ramp_stop_attenuation_level_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_stop_attenuation_level = val/4.0
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_stop_attenuation_level_response', 'attenuation=' + str(self.ramp_stop_attenuation_level)))
	
	def ramp_step_size_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x33
		
			DWORD = Attenuation step size in 0.25 dB units.
		"""
		
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		val = val/4.0
		
		if self.__user_defined_callbacks['ramp_step_size_response'] != None:
			self.__user_defined_callbacks['ramp_step_size_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_step_size = val
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_step_size_response', 'step_size=' + str(self.ramp_step_size)))
	
	def ramp_step_size_bidirectional_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x37
		
			DWORD = Attenuation step size in 0.25 dB units.
		"""
		
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		val = val/4.0
		
		if self.__user_defined_callbacks['ramp_step_size_bidirectional_response'] != None:
			self.__user_defined_callbacks['ramp_step_size_bidirectional_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_step_size_bidirectional = val
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_step_size_bidirectional_response', 'step_size=' + str(self.ramp_step_size_bidirectional)))

	def ramp_wait_time_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x33
		
			DWORD = Time to wait at the end of each attenuation ramp before 
			beginning another attenuation ramp when the tram is in continuous 
			operation mode. This time is in 1 millisecond intervals.
		"""
	
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		
		if self.__user_defined_callbacks['ramp_wait_time_response'] != None:
			self.__user_defined_callbacks['ramp_wait_time_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_wait_time = val
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_wait_time_response', 'wait_time=' + str(self.ramp_wait_time)))
	
	def ramp_wait_time_bidirectional_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x39
		
			DWORD = Time to wait at the end of each attenuation ramp before 
			beginning another attenuation ramp when the tram is in continuous 
			operation mode. This time is in 1 millisecond intervals.
		"""
	
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		
		if self.__user_defined_callbacks['ramp_wait_time_bidirectional_response'] != None:
			self.__user_defined_callbacks['ramp_wait_time_bidirectional_response'](self, status_byte, count, byteblock, val)
		
		self.ramp_wait_time_bidirectional = val
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'ramp_wait_time_bidirectional_response', 'wait_time=' + str(self.ramp_wait_time)))
	
	def ramp_mode_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x33
		
			Byte = 01 for single ramp, and 02 for continuous ramps.
		"""
		
		self.ramp_mode = byteblock[0]
	
	def maximum_attenuation_response(self, status_byte, count, byteblock, debug=False):
	
		"""	Status byte = 0x35
		
			Byte = The maximum attenuation level provided by the attenuator.
		"""
		
		val = (byteblock[3] << 24) + (byteblock[2] << 16) + (byteblock[1] << 8) + byteblock[0]
		val = val/4.0
			
		if self.__user_defined_callbacks['maximum_attenuation_response'] != None:
			self.__user_defined_callbacks['maximum_attenuation_response'](self, status_byte, count, byteblock, val)
		
		self.maximum_attenuation = val/4.0
		
		if debug:
			self.__log.write(self.__cmd_log_string % ( datetime.now(), 'maximum_attenuation_response', 'attenuation=' + str(self.maximum_attenuation)))
		
	#################################################################################
	#								PERIODIC STATUS									#
	#################################################################################
	
	def status_report(self, status_byte, count, byteblock, debug=False):
		
		"""	Status byte = 0x0E
		
			byteblock[0-3] - Reserved
			
			byteblock[4] : dev_status
			
			1 1 1 X 1 1 1 1 	
			|--------------- PLL lock status bit
			  |------------- A parameter was set since the last "Save Settings" command
			    |----------- A command completed
				  |--------- NOT USED!
			        |------- The RF HW is on
			          |----- 0 for sweep up, 1 for sweep down
			            |--- 1 for continuous sweeping
			              |- 1 for single sweep
			        
			byteblock[5] - Current Attenuation as per get_attenuation
		"""

		val = byteblock[5] / 4.0
		
		if debug:
			self.__log.write(self.__status_log_string % ( datetime.now(), 'status_report', status_byte, byteblock[1], val))
		
		if val != self.attenuation_level:
			if debug:
				self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'attenuation_level_change', val))
		
			if self.__user_defined_callbacks['attenuation_level_change'] != None:
				self.__user_defined_callbacks['attenuation_level_change'](self, status_byte, count, byteblock, val)
		
		self.attenuation_level = val
		
		#PLL Lock
		if (byteblock[4] >> 7) & 0x01 == 1:
			val = 1
			
			if val != self.pll_lock:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'pll_lock_change', val))
			
				if self.__user_defined_callbacks['pll_lock_change'] != None:
					self.__user_defined_callbacks['pll_lock_change'](self, status_byte, count, byteblock, val)
			
			self.pll_lock = val
		elif (byteblock[4] >> 7) & 0x01 == 0:
			val = 0
			
			if val != self.pll_lock:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'pll_lock_change', val))

				if self.__user_defined_callbacks['pll_lock_change'] != None:
					self.__user_defined_callbacks['pll_lock_change'](self, status_byte, count, byteblock, val)
			
			self.pll_lock = val
		
		#Save Settings
		if (byteblock[4] >> 6) & 0x01 == 1:
			val = 1
			
			if val != self.save_required:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'save_required_change', val))

				if self.__user_defined_callbacks['save_required_change'] != None:
					self.__user_defined_callbacks['save_required_change'](self, status_byte, count, byteblock, val)
		
			self.save_required = val
		elif (byteblock[4] >> 6) & 0x01 == 0:
			val = 0
			
			if val != self.save_required:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'save_required_change', val))

				if self.__user_defined_callbacks['save_required_change'] != None:
					self.__user_defined_callbacks['save_required_change'](self, status_byte, count, byteblock, val)
		
			self.save_required = val
		
		#Command Complete
		if (byteblock[4] >> 5) & 0x01 == 1:
			val = 0 #Should something be encoded in val here?
			if debug:
				self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'command_complete', val))

			if self.__user_defined_callbacks['command_complete'] != None:
				self.__user_defined_callbacks['command_complete'](self, status_byte, count, byteblock, val)
		elif (byteblock[1] >> 5) & 0x01 == 0:
			pass
		
		#Bit 5 in dev_status not used.
		
		#RF hardware on
		if (byteblock[4] >> 3) & 0x01 == 1:
			val = 1
			
			if val != self.rf_hardware_state:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'rf_hardware_state_change', val))

				if self.__user_defined_callbacks['rf_hardware_state_change'] != None:
					self.__user_defined_callbacks['rf_hardware_state_change'](self, status_byte, count, byteblock, val)
			
			self.rf_hardware_state = val
		elif (byteblock[4] >> 3) & 0x01 == 0:
			val = 0
			
			if val != self.rf_hardware_state:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'rf_hardware_state_change', val))

				if self.__user_defined_callbacks['rf_hardware_state_change'] != None:
					self.__user_defined_callbacks['rf_hardware_state_change'](self, status_byte, count, byteblock, val)
			
			self.rf_hardware_state = val
			
		#Ramp status
		if (byteblock[4] >> 2) & 0x01 == 1:
			#Ramping down
			val = 1
			
			if val != self.ramp_status:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'ramp_status_change', val))

			if self.__user_defined_callbacks['ramp_state_change'] != None:
				self.__user_defined_callbacks['ramp_state_change'](self, status_byte, count, byteblock, val)
			
			self.ramp_status = val
		elif (byteblock[4] >> 2) & 0x01 == 0:
			#Ramping up
			val = 0
			
			if val != self.ramp_status:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'ramp_status_change', val))

			if self.__user_defined_callbacks['ramp_state_change'] != None:
				self.__user_defined_callbacks['ramp_state_change'](self, status_byte, count, byteblock, val)
			
			self.ramp_status = val
		
		#Sweep mode continuous
		if (byteblock[4] >> 1) & 0x01 == 1:
			val = 2
			
			if val != self.ramp_mode:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'ramp_mode_change', val))

			if self.__user_defined_callbacks['ramp_mode_change'] != None:
				self.__user_defined_callbacks['ramp_mode_change'](self, status_byte, count, byteblock, val)
			
			self.ramp_mode = val
			
		elif (byteblock[4] >> 1) | 0x01 == 0:
			pass
		
		#Sweep mode single
		if (byteblock[4]) & 0x01 == 1:
			val = 1
			
			if val != self.ramp_mode:
				if debug:
					self.__log.write(self.__res_log_string % ( datetime.now(), 'status_report', 'ramp_mode_change', val))
			
			if self.__user_defined_callbacks['ramp_mode_change'] != None:
				self.__user_defined_callbacks['ramp_mode_change'](self, status_byte, count, byteblock, val)
			
			self.ramp_mode = val
		elif (byteblock[4]) & 0x01 == 0:
			pass
	
	
	#################################################################################
	#								CALLBACK VARIABLES								#
	#################################################################################
	
	
	__responses = 	{ 	
						0x04 : attenuation_level_response,
						0x0d : attenuation_level_response,
						0x33 : ramp_dwell_time_response,
						0x37 : ramp_dwell_time_bidirectional_response,
						0x30 : ramp_start_attenuation_level_response,
						0x31 : ramp_stop_attenuation_level_response,
						0x32 : ramp_step_size_response,
						0x38 : ramp_step_size_bidirectional_response,
						0x36 : ramp_wait_time_response,
						0x39 : ramp_wait_time_bidirectional_response,
						0x09 : ramp_mode_response,
						0x0e : status_report
					}
	
	""" User defined callbacks must be of the format:
			callback_function(Attenuator device, int status_byte, int count, 
				bytearray byteblock)
		NOTE: Callback functions are called PRIO to updating the Attenuator object 
		for get commands.
	"""
		
	__user_defined_callbacks = 	{ 
									'attenuation_level_response' : None,
									'ramp_dwell_time_response' : None,
									'ramp_dwell_time_bidirectional_response' : None,
									'ramp_start_attenuation_level_response' : None,
									'ramp_stop_attenuation_level_response' : None,
									'ramp_step_size_response' : None,
									'ramp_step_size_bidirectional_response' : None,
									'ramp_wait_time_response' : None,
									'ramp_wait_time_bidirectional_response' : None,
									'ramp_mode_response' : None,
									'attenuation_level_change' : None,
									'pll_lock_change' : None,
									'save_required_change' : None,
									'command_complete' : None,
									'rf_hardware_state_change' : None,
									'ramp_state_change' : None,
									'ramp_mode_change' : None
								}

	def register_callback(self, event, func):
		""" Register a callback to be called when an event happens 
			Callback function to be in the format of:
			callback_function(Attenuator device, int status_byte, 
				int count, bytearray byteblock) 
		
			NOTE: Callback functions are called PRIO to updating the Attenuator object 
				for get commands. """

		if event in self.__user_defined_callbacks:
			self.__user_defined_callbacks[event] = func
			return True
		else:
			return False

	def deregister_callback(self, event):
		""" De-register a callback for an event """

		if event in self.__user_defined_callbacks:
			self.__user_defined_callbacks[event] = None
			return True
		else:
			return False 
		
